from pydoc import resolve
import matplotlib.pyplot as plt
from matplotlib.backend_bases import MouseButton
from matplotlib.patches import Rectangle,Circle
import numpy as np
from pkg_resources import ResolutionError

#from MJOLNIR.Data import Viewer3D
from MJOLNIR._interactiveSettings import selectedColor, \
deselectedColor, States, cut1DKkwargs

# draggable rectangle with the animation blit techniques; see
# http://www.scipy.org/Cookbook/Matplotlib/Animations


class DraggableShape(): # pragma: no cover
    lock = None  # only one can be moved at a time
    selectedShape = None # Only one can be selected at one point
    cidkeypress = None

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)


            
class DraggableRectangle(DraggableShape):# pragma: no cover
    
    def __init__(self, rect,line,plottingObject,Cut1DFunction,figure):
        self.rect = rect
        self.line = line
        self.press = None
        self.background = None
        self._selected = False
        self.plottingObject = plottingObject
        self.Cut1DFunction = Cut1DFunction
        self.figure = figure
        
    def remove(self):
        fig = self.rect.figure
        self.disconnect()
        fig = self.rect.figure
        self.rect.set_visible(False)
        self.rect.remove()
        self.line.set_visible(False)
        self.line.remove()
        if self.selected:
            self.figure.selectedDr = None
        idx = self.figure.shapes.index(self)
        del self.figure.shapes[idx]
        fig.canvas.draw()


    @property
    def selected(self):
        return self._selected
    
    @selected.getter
    def selected(self):
        return self._selected
    
    @selected.setter
    def selected(self,value):
        if self._selected == bool(value):
            return
        if value:
            if not self.figure.selectedDr is None:
                self.figure.selectedDr.selected = False
            self.figure.selectedDr = self
            color = selectedColor
        else:
            color = deselectedColor
            
        self.rect.set_edgecolor(color)
        self.line.set_color(color)
        self.rect.figure.canvas.draw()
        self._selected = bool(value)



    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.rect.axes: return
        if self.figure.lock is not None: return
        contains, attrd = self.rect.contains(event)
        if not contains: return
        
        figure = self.plottingObject
        if event.button == MouseButton.RIGHT and figure.drawState == States.INACTIVE:
            self.remove()
            return 
        self.selected = True
        x0, y0 = self.rect.xy
        self.press = x0, y0, event.xdata, event.ydata
        self.figure.lock = self

        # draw everything but the selected rectangle and store the pixel buffer
        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        self.rect.set_animated(True)
        self.line.set_animated(True)
        canvas.draw()
        self.background = canvas.copy_from_bbox(self.rect.axes.bbox)

        # now redraw just the rectangle
        axes.draw_artist(self.rect)

        # and blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if self.figure.lock is not self:
            return
        if event.inaxes != self.rect.axes: return
        x0, y0, xpress, ypress = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        self.rect.set_x(x0+dx)
        self.rect.set_y(y0+dy)

        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        # restore the background region
        canvas.restore_region(self.background)

        
        center = corner2center(*self.rect.get_xy(), self.rect.get_width(), self.rect.angle)
        
        height = self.rect.get_height()
        angle = np.deg2rad(self.rect.angle+90)
        dx,dy = height*np.array([np.cos(angle),np.sin(angle)])
        self.line.set_data([center[0],center[0]+dx],[center[1],center[1]+dy])
        
        # redraw line and rect with the current rectangle
        axes.draw_artist(self.rect)
        axes.draw_artist(self.line)
        # blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_release(self, event):
        'on release we reset the press data'
        if self.figure.lock is not self:
            return

        self.press = None
        self.figure.lock = None

        # turn off the rect animation property and reset the background
        self.rect.set_animated(False)
        self.line.set_animated(False)
        self.background = None

        # redraw the full figure
        self.rect.figure.canvas.draw()


    ## Class method
    def shapeSelected(figure,axes):
        '''Called when figure has chosen this shape but not yet drawn it'''
        if hasattr(axes,'format_coord_old'):
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) + ' [rect]'
            
        else:
            axes.format_coord_old = axes.format_coord
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) + ' [rect]'

    def inactive(figure,axes,event):
        # TODO: check for 3DViewer#print('here')
        #if hasattr(figure,'axes'): # only Viewer3D
        #    if figure.axis != 2:
        #        return figure,axes,event):

        width = 0.0
        height = 0.05
        center = corner2center(event.xdata,event.ydata, width, -180.0)
        figure.newShape = Rectangle(center,width,height,**cut1DKkwargs) # 
        
        axes.add_patch(figure.newShape)
        
        axes.get_figure().canvas.draw()
        def on_move(self,event):
            if event.inaxes:
                rectx,recty = self.newShape.xy
                center = corner2center(rectx,recty, width, self.newShape.angle)
                mousex = event.xdata
                mousey = event.ydata
                dx,dy = center[0]-mousex, center[1]-mousey
                angle = np.arctan2(dy,dx)+np.pi/2
                
                self.newShape.set_xy(center2corner(*center, width, np.rad2deg(angle)))
                self.newShape.angle = np.rad2deg(angle)
                
                self.newShape.set_height(np.linalg.norm([dx,dy]))
                
                
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw just the current rectangle
                axes.draw_artist(self.newShape)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        figure.newShape.set_animated(True)
        axes.get_figure().canvas.draw()
        figure.background = figure.figure.canvas.copy_from_bbox(axes.bbox)

        # now redraw just the rectangle
        axes.draw_artist(figure.newShape)

        # and blit just the redrawn area
        axes.get_figure().canvas.blit(axes.bbox)
        
        figure.drawState = States.INITIAL

        return on_move

    def initial(figure,axes,_):
        center = corner2center(*figure.newShape.get_xy(), figure.newShape.get_width(), figure.newShape.angle)

        height = figure.newShape.get_height()
        angle = np.deg2rad(figure.newShape.angle+90)
        dx,dy = height*np.array([np.cos(angle),np.sin(angle)])
        
        
        figure.line = axes.plot([center[0],center[0]+dx],[center[1],center[1]+dy],cut1DKkwargs['edgecolor'],zorder=cut1DKkwargs['zorder']+1)[0]
                
        def on_move(self,event):
            if event.inaxes:
                
                rectx,recty = self.newShape.xy
                angle = np.deg2rad(self.newShape.angle+90)
                dirvec = np.array([np.cos(angle),np.sin(angle)])
                
                center = corner2center(rectx,recty, self.newShape.get_width(), self.newShape.angle)
                
                mousePoint = np.array([event.xdata-center[0],event.ydata-center[1]])
                
                ortho = mousePoint-np.dot(dirvec,mousePoint)*dirvec
                sign = np.sign(np.dot(ortho,np.array([dirvec[1],-dirvec[0]])))
                
                width = sign*np.linalg.norm(ortho)*2.0
                
                
                self.newShape.set_width(width)
                self.newShape.set_xy(center2corner(*center, width, self.newShape.angle))
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw both line and rectangle
                axes.draw_artist(self.newShape)
                axes.draw_artist(self.line)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        axes.get_figure().canvas.blit(axes.bbox)
        figure.drawState = States.WIDTH
        return on_move

    def width(figure,axes,_):
        
        rect = Rectangle(figure.newShape.get_xy(),width=figure.newShape.get_width(),
                         height=figure.newShape.get_height(),angle=figure.newShape.angle,
                         **cut1DKkwargs)
        
        #Cut1DFunction = lambda dr: figure.cut1DFunction(figure,dr)
        Cut1DFunction = figure.draggableFunctions[figure.draggableShapes.index(DraggableRectangle)]
        
        axes.add_patch(rect)
        dr = DraggableRectangle(rect,figure.line,figure,Cut1DFunction,figure)
        dr.connect()
        dr.selected = True
        
        figure.new = False
        figure.newShape = None
        figure.line = None
        #figure.shapes.append(rect)
        figure.drawState = States.INACTIVE
       
        

        figure.shapes.append(dr)

        return None

def corner2center(x,y,width,angle):# pragma: no cover
    # Rotate from rectangle space to real space
    angleRad = np.deg2rad(angle+90)
    # Unit vector along height of rectangle
    dirvec = np.array([np.cos(angleRad),np.sin(angleRad)])
    # Orthogonal with half width
    ortho = np.array([dirvec[1],-dirvec[0]])*width*0.5
    # center point is orthogonal to height and half the width
    center = np.array([x,y])+ortho
    return center

def center2corner(x,y,width,angle):# pragma: no cover
    # Rotate from rectangle space to real space
    angleRad = np.deg2rad(angle+90)
    # Unit vector along height of rectangle
    dirvec = np.array([np.cos(angleRad),np.sin(angleRad)])
    # Orthogonal with half width
    ortho = np.array([dirvec[1],-dirvec[0]])*width*0.5
    # corner point is orthogonal to height and half the width but reversed
    corner = np.array([x,y])-ortho
    return corner

def extractCut1DPropertiesRectangle(rect,sample=None, rounding = 4):# pragma: no cover
    # Extract parameters from a rectangle
    xy = rect.get_xy()
    width = rect.get_width()
    height = rect.get_height()
    angle = rect.angle
    
    angleRad = np.deg2rad(angle+90)
    # Unit vector along height of rectangle
    dirvec = np.array([np.cos(angleRad),np.sin(angleRad)])
    ortho = np.array([dirvec[1],-dirvec[0]])
    botCenter = np.array(xy)+ortho*0.5*width
    topCenter = np.array(xy)+ortho*0.5*width+dirvec*height
    
    if not sample is None:
        botCenter = sample.calculateQxQyToHKL(*botCenter)
        topCenter = sample.calculateQxQyToHKL(*topCenter)
    
    minPixel = 0.05
    rlu = not sample is None # RLU only when a sample is provided
    constantBins = False
    ufit = False
    params = {'q1':np.array([np.round(x,rounding) for x in botCenter]),
              'q2':np.array([np.round(x,rounding) for x in topCenter]),
              'width':np.round(np.abs(width),rounding),
              'minPixel':np.round(minPixel,rounding),
              'rlu':rlu,
              'constantBins':constantBins,
              'ufit':ufit}
    return params


def extractCut1DPropertiesRectanglePerpendicular(rect,sample=None, rounding = 4):# pragma: no cover
    # Extract parameters from a rectangle
    xy = rect.get_xy()
    width = rect.get_width()
    height = rect.get_height()
    center = np.asarray(xy)+0.5*np.array([width,height])
    
    rlu = not sample is None # RLU only when a sample is provided
    constantBins = False
    ufit = False
    params = {'center':np.array([np.round(x,rounding) for x in center]),
              'Emin':np.round(xy[1],rounding),
              'Emax':np.round(xy[1]+height,rounding),
              'width':np.round(width,rounding),
              'rlu':rlu,
              'constantBins':constantBins,
              'ufit':ufit}
    return params

def extractCut1DPropertiesRectangleHorizontal(rect,sample=None, rounding = 4):# pragma: no cover
    # Extract parameters from a rectangle
    xy = rect.get_xy()
    width = rect.get_width()
    height = rect.get_height()
    q1 = np.asarray(xy)+np.array([0.0,0.5*height])
    q2 = np.asarray(xy)+np.array([width,height])
    
    
    rlu = not sample is None # RLU only when a sample is provided
    constantBins = False
    ufit = False
    params = {'q1':np.array([np.round(x,rounding) for x in q1]),
              'q2':np.array([np.round(x,rounding) for x in q2]),
              'Emin':np.round(xy[1],rounding),
              'Emax':np.round(xy[1]+height,rounding),
              'rlu':rlu,
              'constantBins':constantBins,
              'ufit':ufit}
    return params

def extractCut1DPropertiesRectangleVertical(rect,sample=None, rounding = 4):# pragma: no cover
    # Extract parameters from a rectangle
    xy = rect.get_xy()
    width = rect.get_width()
    height = rect.get_height()
    q1 = np.asarray(xy)+np.array([0.5*width,0.0])
    
    Energies = [xy[1],xy[1]+height]
    EMin,EMax = np.min(Energies),np.max(Energies)
    
    rlu = not sample is None # RLU only when a sample is provided
    constantBins = False
    ufit = False
    params = {'q':np.round(q1[0],rounding),
              
              'E1':np.round(EMin,rounding),
              'E2':np.round(EMax,rounding),
              'rlu':rlu,
              'constantBins':constantBins,
              'ufit':ufit}
    return params

def extractCut1DPropertiesCircle(circ,sample=None, rounding = 4):# pragma: no cover
    # Extract parameters from a rectangle
    center = circ.center
    width = circ.radius*2
    
    if not sample is None:
        center = sample.calculateQxQyToHKL(*center)
    
    rlu = not sample is None # RLU only when a sample is provided
    constantBins = False
    ufit = False
    params = {'q':np.array([np.round(x,rounding) for x in center]),
              'width':np.round(np.abs(width),rounding),
              'rlu':rlu,
              'constantBins':constantBins,
              'ufit':ufit}
    return params


## Utility functions to make plots interactive

def clearBoxes(self):# pragma: no cover
    """Clear all generated draggable rectangles"""
    if hasattr(self,'ax'):
        axes = self.ax
    else:
        axes = self

    for dr in self.drs:
        dr.remove()
        
    self.drs = []
    axes.get_figure().canvas.draw()


class DraggableCircle(DraggableShape):# pragma: no cover
    
    def __init__(self, circ,plottingObject,Cut1DFunction,figure):
        self.circ = circ
        self.press = None
        self.background = None
        self._selected = False
        self.plottingObject = plottingObject
        self.Cut1DFunction = Cut1DFunction
        self.figure = figure
        
    def remove(self):
        fig = self.circ.figure
        self.disconnect()
        self.circ.set_visible(False)
        self.circ.remove()
        del self.circ
        if self.selected:
            self.figure.selectedDr = None
        idx = self.figure.shapes.index(self)
        del self.figure.shapes[idx]
        fig.canvas.draw()        

    @property
    def selected(self):
        return self._selected
    
    @selected.getter
    def selected(self):
        return self._selected
    
    @selected.setter
    def selected(self,value):
        if self._selected == bool(value):
            return
        if value:
            if not self.figure.selectedDr is None:
                self.figure.selectedDr.selected = False
            self.figure.selectedDr = self
            color = selectedColor
        else:
            color = deselectedColor
            
        self.circ.set_edgecolor(color)

        self.circ.figure.canvas.draw()
        self._selected = bool(value)

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.circ.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.circ.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.circ.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)


    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.circ.axes: return
        if self.figure.lock is not None: return
        contains, attrd = self.circ.contains(event)
        if not contains: return
        
        figure = self.plottingObject
        if event.button == MouseButton.RIGHT and figure.drawState == States.INACTIVE:
            self.remove()
            return 

        self.selected = True
        x0, y0 = self.circ.center
        self.press = x0, y0, event.xdata, event.ydata
        self.figure.lock = self

        # draw everything but the selected rectangle and store the pixel buffer
        canvas = self.circ.figure.canvas
        axes = self.circ.axes
        self.circ.set_animated(True)
        canvas.draw()
        self.background = canvas.copy_from_bbox(self.circ.axes.bbox)

        # now redraw just the rectangle
        axes.draw_artist(self.circ)

        # and blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if self.figure.lock is not self:
            return
        if event.inaxes != self.circ.axes: return
        x0, y0, xpress, ypress = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        self.circ.center = [x0+dx,y0+dy]

        canvas = self.circ.figure.canvas
        axes = self.circ.axes
        # restore the background region
        canvas.restore_region(self.background)


        # redraw line and rect with the current rectangle
        axes.draw_artist(self.circ)
        # blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_release(self, event):
        'on release we reset the press data'
        if self.figure.lock is not self:
            return

        self.press = None
        self.figure.lock = None

        # turn off the rect animation property and reset the background
        self.circ.set_animated(False)
        self.background = None

        # redraw the full figure
        self.circ.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.circ.figure.canvas.mpl_disconnect(self.cidpress)
        self.circ.figure.canvas.mpl_disconnect(self.cidrelease)
        self.circ.figure.canvas.mpl_disconnect(self.cidmotion)

            
            
    ### Class methods
    def shapeSelected(figure,axes):
        '''Called when figure has chosen this shape but not yet drawn it'''
        if hasattr(axes,'format_coord_old'):
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) +' [circle ⊥]'
        else:
            axes.format_coord_old = axes.format_coord
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) + ' [circle ⊥]'
        
    
    def inactive(figure,axes,event):
        radius = 0.01

        figure.newShape = Circle([event.xdata,event.ydata],radius,**cut1DKkwargs) # 
        
        axes.add_patch(figure.newShape)
        
        figure.background = axes.get_figure().canvas.copy_from_bbox(axes.bbox)

        
        figure.cidmove = None
        figure.drawState = States.WIDTH
        
        def on_move(self,event):
            if event.inaxes:
                
                center = self.newShape.center
                mousePoint = np.array([event.xdata-center[0],event.ydata-center[1]])
                
                
                self.newShape.set_radius(np.linalg.norm(mousePoint))
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw both line and rectangle
                axes.draw_artist(self.newShape)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        return on_move
    
    def width(figure,axes,_):
        figure.drawState = States.INACTIVE
        center,radius = figure.newShape.center,figure.newShape.get_radius()
        
        circ = Circle(center,radius,
                        **cut1DKkwargs)
        
        # Find corresponding function to generated DR
        func = figure.draggableFunctions[figure.draggableShapes.index(DraggableCircle)]
        Cut1DFunction = func
        
        axes.add_patch(circ)
        dr = DraggableCircle(circ,figure,Cut1DFunction,figure)
        figure.new = False
        figure.newShape.remove()
        figure.newShape = None

        dr.selected = True
        dr.connect()
        figure.shapes.append(dr)
        
        return None

class DraggableRectanglePerpendicular(DraggableShape):# pragma: no cover
    ### To be used in a QE plane only allowing a perpendicular cut, i.e. Q_perp (integrating over E and Q_para)
    def __init__(self, rect,plottingObject,Cut1DFunction,figure):
        self.rect = rect
        self.press = None
        self.background = None
        self._selected = False
        self.plottingObject = plottingObject
        self.Cut1DFunction = Cut1DFunction
        self.figure = figure
        
    def remove(self):
        fig = self.rect.figure
        self.disconnect()
        self.rect.set_visible(False)
        self.rect.remove()
        del self.rect
        if self.selected:
            self.figure.selectedDr = None
        idx = self.figure.shapes.index(self)
        del self.figure.shapes[idx]
        fig.canvas.draw()        
        fig.ax.redraw_in_frame()

    @property
    def selected(self):
        return self._selected
    
    @selected.getter
    def selected(self):
        return self._selected
    
    @selected.setter
    def selected(self,value):
        if self._selected == bool(value):
            return
        if value:
            if not self.figure.selectedDr is None:
                self.figure.selectedDr.selected = False
            self.figure.selectedDr = self
            color = selectedColor
        else:
            color = deselectedColor
            
        self.rect.set_edgecolor(color)

        self.rect.figure.canvas.draw()
        self._selected = bool(value)

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)


    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.rect.axes: return
        if self.figure.lock is not None: return
        contains, attrd = self.rect.contains(event)
        if not contains: return
        
        figure = self.plottingObject
        if event.button == MouseButton.RIGHT and figure.drawState == States.INACTIVE:
            self.remove()
            return 

        self.selected = True
        x0, y0 = self.rect.xy
        self.press = x0, y0, event.xdata, event.ydata
        self.figure.lock = self

        # draw everything but the selected rectangle and store the pixel buffer
        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        self.rect.set_animated(True)
        canvas.draw()
        self.background = canvas.copy_from_bbox(self.rect.axes.bbox)

        # now redraw just the rectangle
        axes.draw_artist(self.rect)

        # and blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if self.figure.lock is not self:
            return
        if event.inaxes != self.rect.axes: return
        x0, y0, xpress, ypress = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        self.rect.set_xy([x0+dx,y0+dy])

        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        # restore the background region
        canvas.restore_region(self.background)


        # redraw line and rect with the current rectangle
        axes.draw_artist(self.rect)
        # blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_release(self, event):
        'on release we reset the press data'
        if self.figure.lock is not self:
            return

        self.press = None
        self.figure.lock = None

        # turn off the rect animation property and reset the background
        self.rect.set_animated(False)
        self.background = None

        # redraw the full figure
        self.rect.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)

            
            
    ### Class methods
    def shapeSelected(figure,axes):
        '''Called when figure has chosen this shape but not yet drawn it'''
        if hasattr(axes,'format_coord_old'):
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) +' [rect ⊥]'
        else:
            axes.format_coord_old = axes.format_coord
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) + ' [rect ⊥]'
        
    
    def inactive(figure,axes,event):
        width = height = 0.01

        center = [event.xdata,event.ydata]
        lowerLeft = [center[0]-0.5*width, center[1]-0.5*height]
        figure.newShape = Rectangle(lowerLeft,width=width,height=height,**cut1DKkwargs) # 
        
        axes.add_patch(figure.newShape)
        
        figure.background = axes.get_figure().canvas.copy_from_bbox(axes.bbox)

        
        figure.cidmove = None
        figure.drawState = States.WIDTH
        
        def on_move(self,event):
            if event.inaxes:
                
                width = self.newShape.get_width()
                height = self.newShape.get_height()
                center = np.asarray(self.newShape.xy)+np.array([0.5*width,0.5*height])
                
                mousePoint = np.array([event.xdata-center[0],event.ydata-center[1]])
                
                
                newWidth, newHeight = np.abs(mousePoint)*2.0
                lowerLeft = [center[0]-0.5*newWidth, center[1]-0.5*newHeight]

                self.newShape.set_width(newWidth)
                self.newShape.set_height(newHeight)
                self.newShape.set_xy(lowerLeft)
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw both line and rectangle
                axes.draw_artist(self.newShape)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        return on_move
    
    def width(figure,axes,_):
        figure.drawState = States.INACTIVE
        xy,width,height = figure.newShape.xy,figure.newShape.get_width(),figure.newShape.get_height()
        
        rect = Rectangle(xy,width = width, height=height,
                        **cut1DKkwargs)
        
        # Find corresponding function to generated DR
        func = figure.draggableFunctions[figure.draggableShapes.index(DraggableRectanglePerpendicular)]
        Cut1DFunction = func
        
        axes.add_patch(rect)
        dr = DraggableRectanglePerpendicular(rect,figure,Cut1DFunction,figure)
        figure.new = False
        figure.newShape.remove()
        figure.newShape = None

        dr.selected = True
        dr.connect()
        figure.shapes.append(dr)
        
        return None


class DraggableRectangleHorizontal(DraggableShape):# pragma: no cover
    ### To be used in a QE plane only allowing a QCut for constant energy
    def __init__(self, rect,line,plottingObject,Cut1DFunction,figure):
        self.rect = rect
        self.press = None
        self.background = None
        self.line = line
        self._selected = False
        self.plottingObject = plottingObject
        self.Cut1DFunction = Cut1DFunction
        self.figure = figure
        
    def remove(self):
        fig = self.rect.figure
        self.disconnect()
        fig = self.rect.figure
        self.rect.set_visible(False)
        self.rect.remove()
        self.line.set_visible(False)
        self.line.remove()
        if self.selected:
            self.figure.selectedDr = None
        idx = self.figure.shapes.index(self)
        del self.figure.shapes[idx]
        fig.canvas.draw()     

    @property
    def selected(self):
        return self._selected
    
    @selected.getter
    def selected(self):
        return self._selected
    
    @selected.setter
    def selected(self,value):
        if self._selected == bool(value):
            return
        if value:
            if not self.figure.selectedDr is None:
                self.figure.selectedDr.selected = False
            self.figure.selectedDr = self
            color = selectedColor
        else:
            color = deselectedColor
            
        self.rect.set_edgecolor(color)
        self.line.set_color(color)

        self.rect.figure.canvas.draw()
        self._selected = bool(value)

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)


    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.rect.axes: return
        if self.figure.lock is not None: return
        contains, attrd = self.rect.contains(event)
        if not contains: return
        
        figure = self.plottingObject
        if event.button == MouseButton.RIGHT and figure.drawState == States.INACTIVE:
            self.remove()
            return 

        self.selected = True
        x0, y0 = self.rect.xy
        lineXData = self.line.get_xdata()
        lineYData = self.line.get_ydata()
        self.press = x0, y0, event.xdata, event.ydata, lineXData, lineYData
        self.figure.lock = self

        # draw everything but the selected rectangle and store the pixel buffer
        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        self.rect.set_animated(True)
        self.line.set_animated(True)
        canvas.draw()
        self.background = canvas.copy_from_bbox(self.rect.axes.bbox)

        # now redraw just the rectangle
        axes.draw_artist(self.rect)

        # and blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if self.figure.lock is not self:
            return
        if event.inaxes != self.rect.axes: return
        x0, y0, xpress, ypress, lineXData, lineYData = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        self.rect.set_xy([x0+dx,y0+dy])
        
        self.line.set_xdata(lineXData+dx)
        self.line.set_ydata(lineYData+dy)

        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        # restore the background region
        canvas.restore_region(self.background)


        # redraw line and rect with the current rectangle
        axes.draw_artist(self.rect)
        axes.draw_artist(self.line)
        # blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_release(self, event):
        'on release we reset the press data'
        if self.figure.lock is not self:
            return

        self.press = None
        self.figure.lock = None

        # turn off the rect animation property and reset the background
        self.rect.set_animated(False)
        self.line.set_animated(False)
        self.background = None

        # redraw the full figure
        self.rect.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)

            
            
    ### Class methods
    def shapeSelected(figure,axes):
        '''Called when figure has chosen this shape but not yet drawn it'''
        if hasattr(axes,'format_coord_old'):
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) +' [rect constant E]'
        else:
            axes.format_coord_old = axes.format_coord
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) + ' [rect constant E]'
        
    
    def inactive(figure,axes,event):
        width = height = 0.0

        center = [event.xdata,event.ydata]
        lowerLeft = [center[0]-0.5*width, center[1]-0.5*height]
        figure.newShape = Rectangle(lowerLeft,width=width,height=height,**cut1DKkwargs) # 
        
        axes.add_patch(figure.newShape)

        
        
        figure.background = axes.get_figure().canvas.copy_from_bbox(axes.bbox)

        
        figure.cidmove = None
        figure.drawState = States.DIRECTION
        
        def on_move(self,event):
            if event.inaxes:
                
                
                # New width is given by distance from mouse to the left corner
                newWidth = event.xdata-self.newShape.get_x()
                

                self.newShape.set_width(newWidth)
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw both line and rectangle
                axes.draw_artist(self.newShape)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        return on_move
    
    def direction(figure,axes,event):
        
        newWidth = event.xdata-figure.newShape.get_x()
        figure.newShape.set_width(newWidth)

        center = figure.newShape.get_xy()+np.array([0.5*figure.newShape.get_height(),0.0])
        dx = figure.newShape.get_width()
        dy = 0.0
        
        figure.line = axes.plot([center[0],center[0]+dx],[center[1],center[1]+dy],cut1DKkwargs['edgecolor'],zorder=cut1DKkwargs['zorder']+1)[0]


        figure.cidmove = None

        
        figure.drawState = States.WIDTH
        
        def on_move(self,event):
            if event.inaxes:
                
                lowerLeft = self.newShape.get_y()
                center = lowerLeft+self.newShape.get_height()*0.5
                # New width is given by distance from mouse to the left corner
                newHeight = np.abs(event.ydata-center)*2.0
                

                self.newShape.set_height(newHeight)
                self.newShape.set_y(center-0.5*newHeight)
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw rectangle
                axes.draw_artist(self.newShape)
                axes.draw_artist(self.line)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        return on_move

    def width(figure,axes,event):
        figure.drawState = States.INACTIVE
        xy,width,height = figure.newShape.xy,figure.newShape.get_width(),figure.newShape.get_height()
        
        rect = Rectangle(xy,width = width, height=height,
                        **cut1DKkwargs)
        
        # Find corresponding function to generated DR
        func = figure.draggableFunctions[figure.draggableShapes.index(DraggableRectangleHorizontal)]
        Cut1DFunction = func
        
        axes.add_patch(rect)
        dr = DraggableRectangleHorizontal(rect,figure.line,figure,Cut1DFunction,figure)
        figure.new = False
        figure.newShape.remove()
        figure.newShape = None

        dr.selected = True
        dr.connect()
        figure.shapes.append(dr)
        
        return None


class DraggableRectangleVertical(DraggableShape):# pragma: no cover
    ### To be used in a QE plane only allowing a QCut for constant q
    def __init__(self, rect,line,plottingObject,Cut1DFunction,figure):
        self.rect = rect
        self.press = None
        self.background = None
        self.line = line
        self._selected = False
        self.plottingObject = plottingObject
        self.Cut1DFunction = Cut1DFunction
        self.figure = figure
        
    def remove(self):
        fig = self.rect.figure
        self.disconnect()
        fig = self.rect.figure
        self.rect.set_visible(False)
        self.rect.remove()
        self.line.set_visible(False)
        self.line.remove()
        if self.selected:
            self.figure.selectedDr = None
        idx = self.figure.shapes.index(self)
        del self.figure.shapes[idx]
        fig.canvas.draw()     

    @property
    def selected(self):
        return self._selected
    
    @selected.getter
    def selected(self):
        return self._selected
    
    @selected.setter
    def selected(self,value):
        if self._selected == bool(value):
            return
        if value:
            if not self.figure.selectedDr is None:
                self.figure.selectedDr.selected = False
            self.figure.selectedDr = self
            color = selectedColor
        else:
            color = deselectedColor
            
        self.rect.set_edgecolor(color)
        self.line.set_color(color)

        self.rect.figure.canvas.draw()
        self._selected = bool(value)

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)


    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.rect.axes: return
        if self.figure.lock is not None: return
        contains, attrd = self.rect.contains(event)
        if not contains: return
        
        figure = self.plottingObject
        if event.button == MouseButton.RIGHT and figure.drawState == States.INACTIVE:
            self.remove()
            return 

        self.selected = True
        x0, y0 = self.rect.xy
        lineXData = self.line.get_xdata()
        lineYData = self.line.get_ydata()
        self.press = x0, y0, event.xdata, event.ydata, lineXData, lineYData
        self.figure.lock = self

        # draw everything but the selected rectangle and store the pixel buffer
        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        self.rect.set_animated(True)
        self.line.set_animated(True)
        canvas.draw()
        self.background = canvas.copy_from_bbox(self.rect.axes.bbox)

        # now redraw just the rectangle
        axes.draw_artist(self.rect)

        # and blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if self.figure.lock is not self:
            return
        if event.inaxes != self.rect.axes: return
        x0, y0, xpress, ypress, lineXData, lineYData = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        self.rect.set_xy([x0+dx,y0+dy])
        
        self.line.set_xdata(lineXData+dx)
        self.line.set_ydata(lineYData+dy)

        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        # restore the background region
        canvas.restore_region(self.background)


        # redraw line and rect with the current rectangle
        axes.draw_artist(self.rect)
        axes.draw_artist(self.line)
        # blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_release(self, event):
        'on release we reset the press data'
        if self.figure.lock is not self:
            return

        self.press = None
        self.figure.lock = None

        # turn off the rect animation property and reset the background
        self.rect.set_animated(False)
        self.line.set_animated(False)
        self.background = None

        # redraw the full figure
        self.rect.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)

            
            
    ### Class methods
    def shapeSelected(figure,axes):
        '''Called when figure has chosen this shape but not yet drawn it'''
        if hasattr(axes,'format_coord_old'):
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) +' [rect constant Q]'
        else:
            axes.format_coord_old = axes.format_coord
            axes.format_coord = lambda x,y: axes.format_coord_old(x,y) + ' [rect constant Q]'
        
    
    def inactive(figure,axes,event):
        width = height = 0.0

        center = [event.xdata,event.ydata]
        lowerLeft = [center[0]-0.5*width, center[1]-0.5*height]
        figure.newShape = Rectangle(lowerLeft,width=width,height=height,**cut1DKkwargs) # 
        
        axes.add_patch(figure.newShape)

        figure.background = axes.get_figure().canvas.copy_from_bbox(axes.bbox)

        
        figure.cidmove = None
        figure.drawState = States.DIRECTION
        
        def on_move(self,event):
            if event.inaxes:
                
                
                # New width is given by distance from mouse to the left corner
                newHeight = event.ydata-self.newShape.get_y()
                self.newShape.set_height(newHeight)
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw both line and rectangle
                axes.draw_artist(self.newShape)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        return on_move
    
    def direction(figure,axes,event):
        
        newHeight = event.ydata-figure.newShape.get_y()
        figure.newShape.set_height(newHeight)

        center = figure.newShape.get_xy()+np.array([0.5*figure.newShape.get_width(),0.0])
        dx = 0.0
        dy = newHeight
        
        figure.line = axes.plot([center[0],center[0]+dx],[center[1],center[1]+dy],cut1DKkwargs['edgecolor'],zorder=cut1DKkwargs['zorder']+1)[0]

        figure.cidmove = None

        
        figure.drawState = States.WIDTH
        
        def on_move(self,event):
            if event.inaxes:
                
                lowerLeft = self.newShape.get_x()
                center = lowerLeft+self.newShape.get_width()*0.5
                # New width is given by distance from mouse to the left corner
                newWidth = np.abs(event.xdata-center)*2.0
                

                self.newShape.set_width(newWidth)
                self.newShape.set_x(center-0.5*newWidth)
                
                canvas = axes.get_figure().canvas
                # restore the background region
                canvas.restore_region(self.background)
        
                # redraw rectangle
                axes.draw_artist(self.newShape)
                axes.draw_artist(self.line)
        
                # blit just the redrawn area
                canvas.blit(axes.bbox)
                
        return on_move

    def width(figure,axes,event):
        figure.drawState = States.INACTIVE
        xy,width,height = figure.newShape.xy,figure.newShape.get_width(),figure.newShape.get_height()
        
        rect = Rectangle(xy,width = width, height=height,
                        **cut1DKkwargs)
        
        # Find corresponding function to generated DR
        func = figure.draggableFunctions[figure.draggableShapes.index(DraggableRectangleVertical)]
        Cut1DFunction = func
        
        axes.add_patch(rect)
        dr = DraggableRectangleVertical(rect,figure.line,figure,Cut1DFunction,figure)
        figure.new = False
        figure.newShape.remove()
        figure.newShape = None

        dr.selected = True
        dr.connect()
        figure.shapes.append(dr)
        
        return None


# def cancel(self,axes):# pragma: no cover
#     self.resetFormatCoord()
#     self.newShape = None
#     self.new = False
    
#     # Only when something has been drawn is cid set
#     if not self.cidmove is None:
#         axes.get_figure().canvas.mpl_disconnect(self.cidmove)
#         self.cidmove = None
        
#         del self.patches[-1]
#         self.get_figure().canvas.draw_idle()

#     # Reset state and return
#     self.drawState = States.INACTIVE
#     return 



# def on_key_press(self,event):# pragma: no cover
#     if not hasattr(self,'drawState'):
#         return
#     if hasattr(self,'ax'):
#         axes = self.ax
#     else:
#         axes = self
#     if not event.key in cut1DSettingsAll:
#         self.resetFormatCoord()
#         self.suppressPrint = False

#         return
#     if event.key in cut1DSettings['move']:
#         if hasattr(self,'ax'):
#             axes = self.ax
#         else:
#             axes = self
            
#         cancel(self,axes)
            
#         if self.lock is None:
#             self.lock = True
#             self.resetFormatCoord()
#         else:
#             self.lock = None
#             self.format_coord_old = self.format_coord
#             self.format_coord = lambda x,y:self.format_coord_old(x,y)+' [manage]'
#             self.suppressPrint = True
        

#         return
#     elif event.key in cut1DSettings['cut'] and self.new is False:
#         self.lock=True
#         self.resetFormatCoord()
#         dr = self.selectedDr
#         if not dr is None:
#             dr.Cut1DFunction(dr)

#             return
    
#     elif event.key in cut1DSettings['new']:
#         self.lock=True
#         if hasattr(self,'ax'):
#             axes = self.ax
#         else:
#             axes = self
            
#         if not self.drawState == States.INACTIVE:
#             cancel(self,axes)
#             return

            
#         # If new is False, no new shapes will be initialized
#         # If new is not False, the shape is the draggableShape with index self.new-1
#         # Modulus ensures that there is an overflow back to shape 0 instead of out of bounce
#         self.new = np.mod(self.new+1,len(self.draggableShapes)+1)#
        
        
        
#         if self.new>0:
#             self.selectedShape = self.draggableShapes[self.new-1]
#             self.selectedShape.shapeSelected(self,axes)
#             self.suppressPrint = True
#         else:
#             self.resetFormatCoord()
        
#         self.lock = True



# def on_press(self,event):# pragma: no cover
#     if not hasattr(self,'drawState'):
#         return
#     if hasattr(self,'ax'):
#         axes = self.ax
#     else:
#         axes = self

#     if not self.new > 0:
#         return
#     if event.inaxes != axes:
#         return

    
#     # Right-clicking cancels any ongoing action and reset to inactive
#     if event.button == MouseButton.RIGHT:
#         cancel(self,axes)
#         return 


#     if not self.drawState is None:
#         if not self.cidmove is None: # If exists, disconnect it
#             axes.get_figure().canvas.mpl_disconnect(self.cidmove)
#             self.cidmove = None
#         stateName = self.drawState.name.lower()
#         # call the designated state function on the selected shape
#         on_move = getattr(self.selectedShape,stateName)(self,axes,event)
#         self.suppressPrint = True
#         # connect the returned move function if not None. None is return in the last function 
        
#         if not on_move is None:
#             self.cidmove = axes.get_figure().canvas.mpl_connect('motion_notify_event',lambda event:on_move(self,event))
#         else:
#             self.resetFormatCoord()
        
        
# def reset(self):# pragma: no cover
#     if not hasattr(self,'drawState'):
#         return
#     self.suppressPrint=False
#     if hasattr(self,'format_coord_old'):
#         self.format_coord = self.format_coord_old
#         del self.format_coord_old


# def prepareInteractiveCutting(ax,draggables, draggableFunctions):# pragma: no cover
#     fig = ax.get_figure()
#     ax.resetFormatCoord = lambda: reset(ax)
#     ax.new=False
#     ax.cidmove = None
#     ax.drawState = States.INACTIVE
    
#     ax.shapes = []
#     ax.drs = []
#     ax.draggableShapes = draggables
#     ax.draggableFunctions = draggableFunctions
#     ax.ax = ax
    
#     ax.selectedDr = None
#     ax.lock = True
    
#     ax.suppressPrint = False
    
#     fig.canvas.mpl_connect(
#         'key_press_event', lambda event:on_key_press(ax,event))
    
#     fig.canvas.mpl_connect(
#                 'button_press_event', lambda event:on_press(ax,event))


# def prepareInteractiveCuttingView3D(self,draggables, draggableFunctions):# pragma: no cover
#     fig = self.ax.get_figure()
#     self.ax.resetFormatCoord = lambda: reset(self.ax)
#     self.ax.new=False
#     self.ax.cidmove = None
#     self.ax.drawState = States.INACTIVE
    
#     self.ax.shapes = []
#     self.ax.drs = []
#     self.ax.draggableShapes = draggables
#     self.ax.draggableFunctions = draggableFunctions
#     self.ax.ax = self.ax
    
#     self.ax.selectedDr = None
#     self.ax.lock = True
    
#     self.ax.suppressPrint = False
    
#     fig.canvas.mpl_connect(
#         'key_press_event', lambda event:on_key_press(self.ax,event))
    
#     fig.canvas.mpl_connect(
#                 'button_press_event', lambda event:on_press(self.ax,event))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 10 10:31:10 2019

@author: lass
"""

import sys

string  = sys.argv[1]
version = [int(x) for x in string.split('-')[1].split('.')[:-2]]
version[-1]+=1
print('.'.join([str(x) for x in version]))




"""
MJOLNIR Module
^^^^^^^^^^^^^^

"""
#import Geometry, Statistic, Data

import sys,os
sys.path.append('.')

__version__ = '1.3.4'
__author__ = 'Jakob Lass'

__multiFLEXXNormalization__ = os.path.join(os.path.dirname(__file__),'CalibrationMultiFLEXX.csv') 
__flatConeNormalization__ = os.path.join(os.path.dirname(__file__),'CalibrationFlatCone.csv')
__CAMEANormalizationBinning1__ = os.path.join(os.path.dirname(__file__), 'Normalization_1.calib')
__CAMEANormalizationBinning2__ = os.path.join(os.path.dirname(__file__), 'Normalization_2.calib')
__CAMEANormalizationBinning3__ = os.path.join(os.path.dirname(__file__), 'Normalization_3.calib')
__CAMEANormalizationBinning4__ = os.path.join(os.path.dirname(__file__), 'Normalization_4.calib')
__CAMEANormalizationBinning5__ = os.path.join(os.path.dirname(__file__), 'Normalization_5.calib')
__CAMEANormalizationBinning6__ = os.path.join(os.path.dirname(__file__), 'Normalization_6.calib')
__CAMEANormalizationBinning7__ = os.path.join(os.path.dirname(__file__), 'Normalization_7.calib')
__CAMEANormalizationBinning8__ = os.path.join(os.path.dirname(__file__), 'Normalization_8.calib')
__bambusNormalization__ = os.path.join(os.path.dirname(__file__),'CalibrationBambus.csv')



from MJOLNIR.Geometry.Instrument import Instrument,prediction
import MJOLNIR.Geometry.Analyser as Analyser
import MJOLNIR.Geometry.Detector as Detector
import MJOLNIR.Geometry.Wedge as Wedge
from MJOLNIR.Data import Sample
import pytest
import numpy as np
import warnings
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

import os

dataPath = 'samlpedata'

def test_Instrument_init():
    Instr = Instrument()

    assert(np.all(Instr.position==(0,0,0)))

    Det = Detector.Detector(position=(1.0,1,0),direction=(1,0,0))
    Ana = Analyser.Analyser(position=(0.5,0,0),direction=(1,0,1))
    
    wedge = Wedge.Wedge(detectors=[Det,Det],analysers=Ana)

    Instr.wedges=[wedge,wedge]

    assert(Instr.settings['Initialized']==False)



def test_Instrument_error():
    
    try:
        Instr = Instrument(fileName='wrongDummyFile.bin')
        assert False
    except ValueError:
        assert True

    Instr = Instrument()

    Ana = Analyser.FlatAnalyser(position=(0.5,0,0),direction=(1,0,1))

    try:
        Instr.wedges=Ana
        assert False
    except AttributeError:
        assert True

    try:
        Instr.wedges=[Ana,Ana]
        assert False
    except AttributeError:
        assert True

    try:
        Instr.append("Wrong object type")
        assert False
    except AttributeError:
        assert True
    
    try:
        Instr.append(["List of",3.0,"wrong objects"])
        assert False
    except AttributeError:
        assert True

    try:
        Instr.settings = {'Name','New dictionary'}
        assert False
    except NotImplementedError:
        return True


def test_Instrument_warnings():
    Instr = Instrument()

    wedge = Wedge.Wedge(position=(0.5,0,0))

    Instr.wedges = wedge

    with warnings.catch_warnings(record=True) as w: # From https://docs.python.org/3.1/library/warnings.html
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        Instr.wedges = wedge
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[0].category, UserWarning)
        assert 'The list of wedges is not empty! Appending new wedges(s)' in str(w[0].message)


def test_Instrument_append():
    Instr = Instrument()

    wedge = Wedge.Wedge(position=(0.5,0,0))

    Instr.append([wedge,wedge])
    Instr.append(wedge)

    assert(len(Instr.wedges)==3)


def test_Instrument_plot():
    Instr = Instrument()

    wedge = Wedge.Wedge(position=(0.5,0,0))

    Det = Detector.TubeDetector1D(position=(1.0,1,0),direction=(1,0,0))
    Ana = Analyser.FlatAnalyser(position=(0.5,0,0),direction=(1,0,1))

    wedge.append([Det,Ana])
    Instr.append(wedge)
    plt.ioff()
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    Instr.plot(ax)

def test_Instrument_Setting(): 
    Instr = Instrument()
    Instr.settings['SettingVersion']=1.0
    assert(Instr.settings['SettingVersion']==1.0)


def test_Instrument_Initialization():
    Instr = Instrument()

    wedge = Wedge.Wedge(position=(0.5,0,0),concept='ManyToMany')
    pixels=33
    split = [12]
    Det = Detector.TubeDetector1D(position=(1.0,1,0),direction=(1,0,0),pixels=pixels,split=split)
    Ana = Analyser.FlatAnalyser(position=(0.5,0,0),direction=(1,0,1))
    

    wedge.append([Det,Det,Ana,Ana,Ana])

    try:
        Instr.initialize()
        assert False
    except ValueError:
        assert True

    try:
        print(Instr.A4)
        assert False
    except RuntimeError:
        assert True

    try:
        print(Instr.Ef)
        assert False
    except RuntimeError:
        assert True

    Instr.append(wedge)
    try:
        Instr.initialize()
        assert False
    except ValueError:
        assert True
    Instr.wedges[0].detectors[0].split = [0,12,20,pixels]
    Instr.initialize()

    assert(len(Instr.A4)==1)
    assert(len(Instr.A4[0])==2)
    assert(len(Instr.A4[0][0])==pixels)
    assert(len(Instr.A4)==len(Instr.Ef))
    assert(len(Instr.A4[0])==len(Instr.Ef[0]))
    assert(len(Instr.A4[0][0])==len(Instr.Ef[0][0]))
    assert(Instr.settings['Initialized']==True)

    try:
        Instr.A4 = []
        assert False
    except NotImplementedError:
        assert True

    try:
        Instr.Ef = []
        assert False
    except NotImplementedError:
        assert True



def test_Instrument_saveload():
    import os
    Instr = Instrument(position=(0,1,0))
    Instr2 = Instrument()

    wedge = Wedge.Wedge(position=(0.5,0,0))

    Det = Detector.TubeDetector1D(position=(1.0,1,0),direction=(1,0,0))
    Ana = Analyser.FlatAnalyser(position=(0.5,0,0),direction=(1,0,1))

    wedge.append([Det,Ana])
    Instr.append(wedge)

    tempFile = 'temp.bin'
    Instr.save(tempFile)
    Instr2.load(tempFile)
    os.remove(tempFile)
    

    assert(Instr==Instr2)



def test_parseXML(): # Improve this test!

    tempFileName = '__temp__.xml'
        
    Instr = Instrument()
    Instr.settings['Author'] = 'Jakob Lass'

    wedge = Wedge.Wedge(position=(0.5,0,0))

    Det = Detector.TubeDetector1D(position=(1.0,1,0),direction=(1,0,0))
    Ana = Analyser.FlatAnalyser(position=(0.5,0,0),direction=(1,0,1))

    wedge.append([Det,Ana])
    Instr.append([wedge,wedge])
    Instr.append(wedge)
    Instr.saveXML(tempFileName)
        
    InstrLoaded = Instrument(fileName=tempFileName)
    os.remove(tempFileName)

    assert(Instr==InstrLoaded) 


def test_XML_errors():

    fileString = ""
    fileString+="<?xml version='1.0'?>"
    fileString+="<Instrument Initialized='False' Author='Jakob Lass' Date ='16/03/18' position='0.0,0.0,0.0'>"
    fileString+="<Wedge position='0.0,0.0,0.0' concept='ManyToMany'>"
    fileString+="<FlatAnalyser direction='0.707,0.0,0.707' d_spacing='3.35' mosaicity='60' width='0.05' height='0.1'></FlatAnalyser>"
    fileString+="<TubeDetector1D position='1.198,0.0580,0.71' direction='0.998,0.04841,0.0' pixels='456' length='0.883' diameter='0.02' split='57, 114, 171, 228, 285, 342, 399'></TubeDetector1D>"
    fileString+="</Wedge>"
    fileString+="</Instrument>"

    temp_file = 'Tempfile.xml'
    f = open(temp_file,'w')
    f.write(fileString)
    f.close()

    try:
        Instr = Instrument(fileName=temp_file)
        del Instr
        assert False
    except ValueError:
        assert True

    fileString = ""
    fileString+="<?xml version='1.0'?>"
    fileString+="<Instrument Initialized='False' Author='Jakob Lass' Date ='16/03/18' position='0.0,0.0,0.0'>"
    fileString+="<Wedge position='0.0,0.0,0.0' concept='ManyToMany'>"
    fileString+="<FlatAnalyser position='0.0580,0.71' direction='0.707,0.0,0.707' d_spacing='3.35' mosaicity='60' width='0.05' height='0.1'></FlatAnalyser>"
    fileString+="<TubeDetector1D position='1.198,0.0580,0.71' direction='0.998,0.04841,0.0' pixels='456' length='0.883' diameter='0.02' split='57, 114, 171, 228, 285, 342, 399'></TubeDetector1D>"
    fileString+="</Wedge>"
    fileString+="</Instrument>"
    f = open(temp_file,'w')
    f.write(fileString)
    f.close()
    try:
        Instr = Instrument(fileName=temp_file)
        assert False
    except AttributeError:
        assert True

    fileString = ""
    fileString+="<?xml version='1.0'?>"
    fileString+="<Instrument Initialized='False' Author='Jakob Lass' Date ='16/03/18' position='0.0,0.0,0.0'>"
    fileString+="<FlatAnalyser position='0.0,0.0,0.0' concept='ManyToMany'>"
    fileString+="<FlatAnalyser position='0.0580,0.71' direction='0.707,0.0,0.707' d_spacing='3.35' mosaicity='60' width='0.05' height='0.1'></FlatAnalyser>"
    fileString+="<TubeDetector1D position='1.198,0.0580,0.71' direction='0.998,0.04841,0.0' pixels='456' length='0.883' diameter='0.02' split='57, 114, 171, 228, 285, 342, 399'></TubeDetector1D>"
    fileString+="</FlatAnalyser>"
    fileString+="</Instrument>"
    f = open(temp_file,'w')
    f.write(fileString)
    f.close()
    try:
        Instr = Instrument(fileName=temp_file)
        assert False
    except ValueError:
        assert True
    os.remove(temp_file)

def test_instrument_string_dummy(): # Todo: Improve test!
    Instr = Instrument()

    string = str(Instr)
    del string
    assert True
    
def test_instrument_create_xml():

    Instr = Instrument()
    filename = 'temp'
    Instr.generateCAMEAXML(filename)

    Instr2 = Instrument(fileName=filename+'.xml')
    os.remove(filename+'.xml')
    assert(len(Instr2.wedges)==8)


@pytest.mark.unit
def test_Normalization_tables(quick):

    Instr = Instrument(fileName=os.path.join('Data','CAMEA_Updated.xml'))
    Instr.initialize()

    NF = os.path.join(dataPath,'camea2023n000083.hdf')
    #AF = 'TestData/1024/A4Normalization.h5'

    try:
        Instr.generateCalibration(Vanadiumdatafile=NF ,savelocation=os.path.join(dataPath,''),plot=False,tables=[]) # No binning specified 
        assert False
    except AttributeError:
        assert True

    try:
        Instr.generateCalibration(Vanadiumdatafile=NF ,savelocation=os.path.join(dataPath,''),plot=False,tables=['Nothing?']) # Wrong binning
        assert False
    except AttributeError:
        assert True

    if not quick==True:
        Instr.generateCalibration(Vanadiumdatafile=NF,  savelocation=os.path.join(dataPath,''),plot=False,tables=[1,3,8],sampleMass=4.7) 
    else:
        Instr.generateCalibration(Vanadiumdatafile=NF ,savelocation=os.path.join(dataPath,''),plot=False,tables=[1],sampleMass=4.7) 


def test_Prediction():
    A3Start = 0.0
    A3Stop = 100
    A3Steps = 101 
    Ei = 5.0
    A4 = [-36,-40]
    points = False
    # [H,K,L,A3,A4,0.0,0.0,Ei,Ef]
    HKL1 = np.array([1,0,0])
    HKL2 = np.array([0,0,1])
    A3R1 = 25.0
    A3R2 = 115.0
    #r1 = np.array([1,0,0,25.0,-24,0.0,0.0,Ei,Ei])
    #r2 = np.array([0,0,1,115.0,-24,0.0,0.0,Ei,Ei])
    
    cell = np.array([6.0,6.0,6.0,90.0,90.0,90.0])

    sample = Sample.calculateSample(cell,HKL1,HKL2,A3R1=A3R1,A3R2=A3R2, Ei=Ei,Ef=Ei)

    plt.ion()

    ax = prediction(A3Start=A3Start,A3Stop=A3Stop,A3Steps=A3Steps,A4Positions=A4,Ei=Ei,sample=sample,
    points=points, instrument='CAMEA')

    ax = prediction(A3Start=A3Start,A3Stop=A3Stop,A3Steps=A3Steps,A4Positions=A4,Ei=Ei,sample=sample,
    points=points, instrument='MultiFLEXX')

    ax = prediction(A3Start=A3Start,A3Stop=A3Stop,A3Steps=A3Steps,A4Positions=A4,Ei=Ei,sample=sample,
    points=points, instrument='Bambus')
    




from __future__ import division
from abc import ABCMeta
import numpy as np
from MJOLNIR.Data import DataSet,DataFile
import sympy
import warnings
import pickle
# Compability of python 2 and 3 with metaclasses
# Python 2 and 3:
from six import with_metaclass
# or
from future.utils import with_metaclass

def requiredArguments(func,requiredNames):
    """Return list of arguments not found in arguments of func"""
    varnames = func.__code__.co_varnames
    
    notInSignature = np.array([None if x in varnames else x for x in requiredNames])
    notInSignature = notInSignature[notInSignature != np.array(None)]
    return notInSignature

class MaskingObjectMeta(ABCMeta):
    """MetaClass to ensure call signatures of derived classes of MaskingObject"""

    def __new__(mcls,name,bases,namespace): # When a new class is created, that has this metaclass
        missingMethods = [] # Hold all methods missing
        wrongSignature = [] # Hold all methods having wrong signature
        
        if not name in ['MaskingObject','MultiMask']: # If MaskingObject, then just continue
            
            if not 'call' in namespace: # Check if call exists
                missingMethods.append('call')
                
            
            if namespace['dimensionality'] == '2D' or namespace['dimensionality'] == '3D':
                if 'plot' in namespace: # Check if plot has the right signature

                    callFunction = namespace['plot']
                    requiredNames = ['ax','transformation']
                    notInSignature = requiredArguments(callFunction,requiredNames)
                    if len(notInSignature)>0:
                        wrongSignature.append(['plot',requiredNames,notInSignature])
                else:
                    missingMethods.append('plot')
            
            # Write error report
            ErrorMessage = []
            if len(missingMethods)>0:
                ErrorMessage.append("Can't instantiate abstract class {} with abstract methods\n".format(name)+', '.join(missingMethods))
            
            for entry in wrongSignature:
                methodName = entry[0]
                requiredNames = entry[1]
                notInSignature = entry[2]
                ErrorMessage.append('The "{}" method on "{}" must have all of the following arguments in call signature [{}]. Following are missing: [{}]'.format(methodName,name,', '.join(requiredNames),', '.join(notInSignature)))
            if len(ErrorMessage)>0:
                raise TypeError('\n'.join(ErrorMessage))
        
        return ABCMeta.__new__(mcls,name,bases,namespace)
    
            
class MaskingObject(with_metaclass(MaskingObjectMeta)):
    """Base class for all masking objects"""
    #dimensionality = '2D'
    def __init__(self,coordinates=None,maskInside=True,name=None):

        self.maskInside = maskInside
        self._coordinates = []
        if coordinates is None:
            self.coordinates = []
        else:
            if isinstance(coordinates,str):
                coordinates = [coordinates]
            self.coordinates = coordinates
        self.bar = self.negate
        if name is None:
            name = 'mask_0'
        self.name = name
        self.negated = False

    def __eq__(self,other):
        if not isinstance(self,(type(other))): # not same type:
            return False
        kwargs = self.generateInputKwargs()
        otherkwargs = other.generateInputKwargs()
        if not np.all([selfkeys in otherkwargs.keys() for selfkeys in kwargs.keys()]):
            return False
        
        for key,value in kwargs.items():
            try:
                truth = np.all(value == otherkwargs[key])
            except ValueError:
                truth = np.all(np.isclose(value,otherkwargs[key]))
            if not truth:
                return False
        return True

    def __hash__(self):
        return hash(self.__dict__.values())

    def plot(self,ax): # pragma: no cover
        """plotting function to put the masked object onto the figure"""
        pass
    
    
    def __call__(self,X=None,*args,**kwargs):
        if X is None:
            return self.call()
        elif isinstance(X,(DataSet.DataSet)):
            result = []
            for x in X:
                temp = self.call(x,*args,**kwargs)
                if not len(temp.shape) == len(x.I.shape):
                    temp = np.outer(temp,np.ones_like(x.I[0],dtype=bool)).reshape(*x.I.shape)
                result.append(temp)
            return result
        else:
            temp = self.call(X,*args,**kwargs)
            #if not len(temp.shape) == len(X.I.shape):
            #    temp = np.outer(temp,np.ones_like(X.I[0],dtype=bool)).reshape(*X.I.shape)
            return temp#self.call(X,*args,**kwargs)
    
    def __add__(self,other):
        return MultiMask(masks=[self,other],operation=np.logical_or)
    
    def __mul__(self,other):
        return MultiMask(masks=[self,other],operation=np.logical_and)
    
    def __neg__(self):
        return self.negate()

    def __pos__(self):
        return self.clone()
    
    def __sub__(self,other):
        other.negated = True
        return self+(-other)
    
    
    def generateInputKwargs(self):
        """Generate dictionary with initial args/kwargs used for init"""
        kwargs = {}
        init = self.__class__.__init__
        initCode = init.__code__
        args = initCode.co_argcount
        for X in initCode.co_varnames[1:args]:
            kwargs[X] = self.__dict__[X]
        return kwargs
        
    def clone(self):
        """Return clone of self"""
        return self.__class__(**self.generateInputKwargs())
        
    def negate(self):
        
        newMask = self.clone()
        newMask.maskInside = not newMask.maskInside
        newMask.negated = True
        return newMask

    def save(self,file):
        mp = pickle.dumps(self)
        correct = True
        if not '.' in file:
            correct = False
        elif not file.split('.')[-1].strip() == 'mask':
            correct = False
        
        if not correct:
            file = file+'.mask'
        with open(file,'wb') as f:
            f.write(mp)
        
class MultiMask(MaskingObject):
    def __init__(self,masks,operation=np.logical_and,negated=False):
        if not hasattr(masks,'__len__'):
            masks = [masks]
            
        self.masks = masks

        names = self.getNames()
        newNames = [names[0]] # first mask always keeps it name
        for idx,name in enumerate(names[1:],start=1):

            if len(names)>idx+1:
                otherNames = newNames+names[idx+1:]
            else:
                otherNames = newNames

            while name in otherNames:
                try:
                    idx = int(name.split('_')[-1])
                except ValueError:
                    name = name+'_1'
                else:
                    name = name[:-(len(str(idx))+1)] + '_' + str(idx+1)
            newNames.append(name)

        for mask,newName in zip(self.getMasks(),newNames):
            mask.name = newName

        self.operation = operation
        self.negated = negated
        self.bar = self.negate
        coordinates = np.concatenate([np.array(m.coordinates) for m in self.masks]).flatten()
        self.coordinates = np.array(list(set(coordinates))) # unique coordinates

    def getNames(self):
        return [mask.name for mask in self.getMasks()]

    def getMasks(self):
        masks = np.concatenate([[mask] if not isinstance(mask,(MultiMask)) else mask.getMasks() for mask in self.masks],axis=0).flatten()
        uniqueMasks = []
        for mask in masks:
            if not mask in uniqueMasks:
                uniqueMasks.append(mask)
       
        return uniqueMasks
        
    
    def call(self,*args,**kwargs):
        masks = []
        for m in self.masks:
            masks.append(m(*args,**kwargs))
        
        results = self.operation(*masks)
        if self.negated:
            results = np.logical_not(results)
        return results
    
    def negate(self):
        newMask = self.clone()
        newMask.negated = not newMask.negated
        return newMask
        
        
    def plot(self,ax,transformation=None,*args,**kwargs):# pragma: no cover
        for m in self.masks:
            m.plot(ax,transformation,*args,**kwargs)
    
    def __add__(self,other):
        return MultiMask(masks=[self,other],operation=np.logical_or)
    
    def __mul__(self,other):
        return MultiMask(masks=[self,other],operation=np.logical_and)
    
    def __neg__(self):
        return self.negate()

    def __pos__(self):
        return self.clone()
    
    def __sub__(self,other):
        return self+(-other)

    
    def generateInputKwargs(self):
        """Generate dictionary with initial args/kwargs used for init"""
        kwargs = {}
        init = self.__class__.__init__
        initCode = init.__code__
        args = initCode.co_argcount
        for X in initCode.co_varnames[1:args]:
            kwargs[X] = self.__dict__[X]
        return kwargs
        
    def clone(self):
        """Return clone of self"""
        return self.__class__(**self.generateInputKwargs())



def RotationMatrix(theta,deg=True):
    """Generate 2D rotation matrix given angle theta"""
    if deg == True:
        theta = np.deg2rad(theta)
    M = np.array([[np.cos(theta),-np.sin(theta)],[np.sin(theta),np.cos(theta)]])
    return M

def RotationMatrix3D(theta,n=None,deg=True):
    """Generate 3D rotation matrix given angle and rotation axis"""
    if deg == True:
        theta = np.deg2rad(theta)
    if n is None:
        M = np.eye((3,3))
    else:
        n = np.array(n,dtype=float)
        n*=1.0/np.linalg.norm(n)
        cost = np.cos(theta)
        sint = np.sin(theta)
        M = np.zeros((3,3))
         # taken from http://scipp.ucsc.edu/~haber/ph216/rotation_12.pdf
        M[0,0] = cost+n[0]**2*(1-cost)
        M[0,1] = n[0]*n[1]*(1-cost)-n[2]*sint
        M[0,2] = n[0]*n[2]*(1-cost)+n[1]*sint
        
        M[1,0] = n[0]*n[1]*(1-cost)+n[2]*sint
        M[1,1] = cost+n[1]**2*(1-cost)
        M[1,2] = n[1]*n[2]*(1-cost)-n[0]*sint
        
        M[2,0] = n[0]*n[2]*(1-cost)-n[1]*sint
        M[2,1] = n[1]*n[2]*(1-cost)+n[0]*sint
        M[2,2] = cost+n[2]**2*(1-cost)
        
    return M

class lineMask(MaskingObject):
    dimensionality = '1D'
    def __init__(self,start,end,maskInside=True,coordinates=None,name=None):
        """Generate a rectangle mask with side corner1 to corner 2 and use corner 3 to define height.
        
        args:
            start (1D point): Starting position
            
            end (1D point): Ending corner
            
        kwargs:
            
            maskInside (bool): If true, points inside is masked otherwise outside (default True)
            
            coordinates (list of str): List containing names of attributes on point object. If None, use x,y (default None)
            
        Example:
            
            >>> P = np.array([[0.5,2.6])
            >>> b = lineMask(*P)
            >>> mask = b(points.T)
            >>> inside = points[mask].T
            
        """
        super(lineMask,self).__init__(coordinates=coordinates,maskInside=maskInside,name=name)
        
        self.start = start
        self.end = end
        
    
        
    
        
        
    def plot(self,ax,transformation=None,*args,**kwargs):# pragma: no cover
        warnings.warn('It is not possible to plot a 1D masks.')#raise NotImplementedError('It is not possible to plot a 1D masks.')
    
    def call(self,x,y=None,z=None):
        if len(self.coordinates)>0:
            points = np.array([getattr(x,coord) for coord in self.coordinates])
        else:
            if y is None:
                #if x.shape[0] != 1:
                #    raise AttributeError('Dimensionality of x is to be (1,N), received {}'.format(x.shape))
                points = [x]
            else:
                raise AttributeError('Dimensionality of input is to be (1,N), received both an x and y')
        
        mask = np.sum([np.logical_and(p>self.start,p<=self.end) for p in points],axis=0)>0
        
        if isinstance(x,DataFile.DataFile):
            if len(mask) == len(x.I) and len(mask.shape)==1: # We have a mask of A3
                    mask = np.outer(mask,np.ones_like(x.I[0],dtype=bool)).reshape(*x.I.shape)
            if len(mask) == len(x.I)+1 and len(mask.shape)==1: # For some reason A3 is 1 longer than I...
                mask = np.outer(mask[:-1],np.ones_like(x.I[0],dtype=bool)).reshape(*x.I.shape)
        
        if self.maskInside == False:
            mask = np.logical_not(mask)
        return mask

class rectangleMask(MaskingObject):
    dimensionality = '2D'
    def __init__(self,corner1,corner2,corner3=None,maskInside=True,coordinates=None,name=None):
        """Generate a rectangle mask with side corner1 to corner 2 and use corner 3 to define height.
        
        args:
            corner1 (2D point): First corner
            
            corner2 (2D point): Second corner
            
        kwargs:
            corner3 (2D point): Third corner used to define height of box. If corner3 is not set corner1 and corner2 are assumed to be opposite corners.
            
            maskInside (bool): If true, points inside is masked otherwise outside (default True)
            
            coordinates (list of str): List containing names of attributes on point object. If None, use x,y (default None)
            
        Example:
            >>> fig,ax = plt.subplots()
            >>> points = np.random.rand(3000,2)
            >>> ax.scatter(points[:,0],points[:,1],color='k')
            >>> P = np.array([[0.15, 0.50],[0.42, 0.16],[0.85, 0.2]])
            >>> b = rectangleMask(*P)
            >>> ax.set_aspect('equal')
            >>> b.plot(ax)
            >>> ax.scatter(*b.center,color='r')
            >>> ax.scatter(P[:,0],P[:,1],color='m')
            >>> mask = b(points.T)
            >>> inside = points[mask].T
            >>> plt.scatter(inside[0],inside[1],color='b')
            >>> plt.scatter(*np.array([b.corners]).T,color='g')
            
        """
        super(rectangleMask,self).__init__(coordinates=coordinates,maskInside=maskInside,name=name)
        
        self.corner1 = np.array(corner1,dtype=float)
        self.corner2 = np.array(corner2,dtype=float)
        
        if isinstance(corner3, np.ndarray):
            if corner3.shape == ():
                self.corner3 = None
            else:
                self.corner3 = corner3
        else:
            if corner3 is None:
                self.corner3 = None
            else:
                self.corner3 = np.array(corner3,dtype=float)
        
        #self.corners = [corner1,corner2]

        
        corner3,corner4 =self.calculateCorners(self.corner1,self.corner2,self.corner3)

        
        self.rotation = np.arctan2(self.edge[1],self.edge[0])
            
        self.rotationMatrix = RotationMatrix(-self.rotation,deg = False)
        
        # Calculate center position 
        self.center = np.mean(self.corners,axis=0).reshape(2,1)
    
    
        
    def calculateCorners(self,c1,c2,c3):
        if c3 is None:
            c3 = c2
            diff = c3-c1
            c2 = c1+np.array([0,diff[1]])
            
        self.edge = c2 - c1 # Defined as line between first and second corner
        self.height = np.linalg.norm(self.edge) # Find side length of rectangle
        self.edge*=1.0/self.height
        self.cross = np.array([-self.edge[1],self.edge[0]]) # Vector orthogonal to edge
        
        self.length = np.dot(self.cross,c3-c1)
        self.lengthSign = np.sign(self.length)
        self.length = np.abs(self.length)
        
        c3 = c2+self.lengthSign*self.cross*self.length/np.linalg.norm(self.cross)
        c4 = c3-c2+c1
        self.corners = [c1,c2,c3,c4]
        return c3,c4
        
        
    def plot(self,ax,transformation=None,*args,**kwargs):# pragma: no cover
        points = np.array([self.corners[0],self.corners[1],self.corners[2],self.corners[3],self.corners[0]]).T
        if not transformation is None:
            points = transformation(*points)
        ax.plot(*points,**kwargs)
    
    def call(self,x,y=None,z=None):
        if len(self.coordinates)>0:
            points = np.array([getattr(x,coord) for coord in self.coordinates])
        else:
            if y is None:
                if x.shape[0] != 2:
                    raise AttributeError('Dimensionality of x is to be (2,N), received {}'.format(x.shape))
                points = np.array(x)
            else:
                points = np.array([x,y])
        # calculate mask by rotating all points so rectangle is along coordinate axes
        # This makes the logical checks easy
        pointShape = points.shape[1:]
        points.shape = (2,-1)
        
        rotatedPoints = np.einsum('ij,jk->ik',self.rotationMatrix,np.array(points)-self.center)
        rotatedPoints.shape = np.concatenate([[2],pointShape])
        logic = np.array([np.logical_and(x>-0.5*edge,x<0.5*edge) for x,edge in zip(rotatedPoints,[self.height,self.length])])
        
        mask = np.all(logic,axis=0)
        if self.maskInside == False:
            mask = np.logical_not(mask)
        return mask
    

class boxMask(MaskingObject):
    dimensionality = '3D'
    def __init__(self,corner1,corner2,corner3,corner4=None,maskInside=True,coordinates=None,name=None):
        """Generate a box mask with side corner1, corner 2, corner 3 and use corner 4 to define height.
        
        args:
            corner1 (3D point): First corner
            
            corner2 (3D point): Second corner
            
            corner2 (3D point): Third corner
            
        kwargs:
            corner4 (3D point): Fourth corner used to define height of box. If corner4 is not set corner1, corner2, corner3 are assumed to be maximal positions.
            
            maskInside (bool): If true, points inside is masked otherwise outside (default True)
            
            coordinates (list of str): List containing names of attributes on point object. If None, use x,y (default None)
            
        
        """
        super(boxMask,self).__init__(coordinates=coordinates,maskInside=maskInside,name=name)
        self.corner1 = np.array(corner1,dtype=float)
        self.corner2 = np.array(corner2,dtype=float)
        self.corner3 = np.array(corner3,dtype=float)
        self.corner4 = np.array(corner4,dtype=float)
        
        

        if corner4 is None: # use provided corners and find bounding box
            X,Y,Z = [[np.min(x),np.max(x)] for x in np.array([corner1,corner2,corner3]).T]
            c1 = np.array([X[0],Y[0],Z[0]])
            c2 = np.array([X[1],Y[0],Z[0]])
            c3 = np.array([X[0],Y[1],Z[0]])
            c4 = np.array([X[1],Y[1],Z[0]])
            c5 = np.array([X[0],Y[0],Z[1]])
            c6 = np.array([X[1],Y[0],Z[1]])
            c7 = np.array([X[0],Y[1],Z[1]])
            c8 = np.array([X[1],Y[1],Z[1]])
            
            self.length = X[1]-X[0]
            self.width = Y[1]-Y[0]
            self.height = Z[1]-Z[0]
            
            self.planeNormal = np.array([0,0,1])#np.cross(edge1,edge2)
            #planeNormal *= 1.0/np.linalg.norm(planeNormal)
            
            #self.rotationMatrix = RotationMatrix3D(0.0,n=None,deg=False)
            
        else:
            c1 = np.array(corner1,dtype=float)
            c2 = np.array(corner2,dtype=float)
            
            # Start by finding common plane for c1, c2, and  c3
            edge1 = c2 - c1 # Defined as line between first and second corner
            edge2 = corner3 - c1 # Defined as line between first and second corner
            
            self.length = np.linalg.norm(edge1)
            
            edge1 *= 1.0/self.length
            edge2 *= 1.0/np.linalg.norm(edge2)
            
            planeNormal = np.cross(edge1,edge2)
            self.planeNormal = planeNormal/np.linalg.norm(planeNormal)
            
            cross = np.cross(edge1,self.planeNormal)
            
            
            width = np.dot(cross,corner3 - c1) # distance from c3 orthogonally on edge1
            widthSign = np.sign(width)
            self.width = np.abs(width)
            
            
            # calculate corner point for c3 and c4
            c4 = c2+widthSign*cross*self.width
            
            c3 = c4-c2+c1 # Find corresponding c4
            
            # find cube
            height = np.dot(self.planeNormal,corner4-c1)
            heightSign = np.sign(height)
            self.height = np.abs(height)
            
            c5 = c1+heightSign*self.planeNormal*self.height
            c6 = c5+c2-c1 
            c7 = c5+c3-c1 
            c8 = c5+c4-c1 
            
        self.corners = np.array([c1,c2,c3,c4,c5,c6,c7,c8])
        self.center = np.array([np.mean(X) for X in self.corners.T]).reshape(1,-1)
        
        # Create rotation matrix.

        # rotate planeNormal to lie along x,z plane (https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d)
        
        v = np.cross(self.planeNormal,np.array([0.0,1.0,0.0]))
        Vs = np.zeros((3,3))
        Vs[0,1] = -v[2]
        Vs[0,2] = +v[1]
        Vs[1,2] = -v[0]
        Vs[1,0] = -Vs[0,1]
        Vs[0,2] = -Vs[0,2]
        Vs[2,1] = -Vs[1,2]
        
        c = np.dot(self.planeNormal,np.array([0.0,1.0,0.0]))
        
        rot1 = np.eye(3)+Vs+np.dot(Vs,Vs)*1.0/(1+c)
        
        
        c1r1,c2r1,c3r1,c4r1,c5r1,c6r1,c7r1,c8r1 = [np.dot(rot1,(x-self.center).T) for x in [c1,c2,c3,c4,c5,c6,c7,c8]]
        
        # rotate around y into plane using edge 1 (c2-c1)
        edge = c2r1-c1r1
        theta = np.arctan2(edge[2],edge[0])
        
        rot2 = RotationMatrix3D(theta,n=[0,1,0],deg=False)
        self.rotationMatrix = np.einsum('ij,jk->ik',rot2,rot1)

        
        
    def plot(self,ax,transformation = None, *args,**kwargs):# pragma: no cover
        
        # Draw square for Z = 0
        points = np.array([self.corners[0],self.corners[1],self.corners[3],self.corners[2],self.corners[0]]).T
        if not transformation is None:
            points = transformation(*points)
        p = ax.plot3D(*points,**kwargs)
        if not 'color' in kwargs:
            color = p[0].get_color()
            kwargs['color'] = color
        
        # Draw square for Z = 1
        points = np.array([self.corners[4],self.corners[5],self.corners[7],self.corners[6],self.corners[4]]).T
        if not transformation is None:
            points = transformation(*points)
        ax.plot3D(*points,**kwargs)
        
        # Draw square for X = 0
        points = np.array([self.corners[0],self.corners[2],self.corners[6],self.corners[4],self.corners[0]]).T
        if not transformation is None:
            points = transformation(*points)
        ax.plot3D(*points,**kwargs)
        # Draw square for X = 1
        points = np.array([self.corners[1],self.corners[3],self.corners[7],self.corners[5],self.corners[1]]).T
        if not transformation is None:
            points = transformation(*points)
        ax.plot3D(*points,**kwargs)
    
    
    def call(self,x,y=None,z=None):
        if len(self.coordinates)>0:
            points = np.array([getattr(x,coord) for coord in self.coordinates])
        else:
            if y is None:
                if x.shape[0] != 3:
                    raise AttributeError('Dimensionality of x is to be (3,N), received {}'.format(x.shape))
                points = x
            else:
                if z is None:
                    raise AttributeError('Recieved only X and Y, but no Z')
                points = np.array([x,y,z])
        # calculate mask by rotating all points so rectangle is along coordinate axes
        # This makes the logical checks easy
        pointShape = points.shape[1:]
        
        points.shape = (3,-1)
        
        rotatedPoints = np.einsum('ij,jk->ik',self.rotationMatrix,np.array(points)-self.center.T)
        rotatedPoints.shape = np.concatenate([[3],pointShape]).astype(int)
        logic = np.array([np.logical_and(x>=-0.5*edge,x<=0.5*edge) for x,edge in zip(rotatedPoints,[self.length,self.height,self.width])])
        
        mask = np.all(logic,axis=0)
        if self.maskInside == False:
            mask = np.logical_not(mask)
        return mask


class circleMask(MaskingObject):
    
    dimensionality = '2D'
    
    def __init__(self,center,radiusPoint=None,radius=None,maskInside=True,coordinates=None,name=None):
        """Generate a circular mask with center at center and edge at radiusPoint.
        
        args:
            center (2d point): center
            
        kwargs:
            radiusPoint (2d point): Circumference goes through this point (default None)

            radius (float): radius if  circle/sphere. If set, it overrides radiusPoint (default none)

            maskInside (bool): If true, points inside is masked otherwise outside (default True)
            
            coordinates (list of str): List containing names of attributes on point object. If None, use x,y (default None)
            
        Example:
            >>> fig,ax = plt.subplots()
            >>> points = np.random.rand(3000,2)
            >>> ax.scatter(points[:,0],points[:,1],color='k')
            >>> P = np.array([[0.15, 0.50],[0.42, 0.16]])
            >>> b = circleMask(*P)
            >>> ax.set_aspect('equal')
            >>> b.plot(ax)
            >>> ax.scatter(*b.center,color='r')
            >>> ax.scatter(P[:,0],P[:,1],color='m')
            >>> mask = b(points.T)
            >>> inside = points[mask].T
            >>> plt.scatter(inside[0],inside[1],color='b')
            >>> plt.scatter(*np.array([b.corners]).T,color='g')
            
        """
        super(circleMask,self).__init__(coordinates=coordinates,maskInside=maskInside,name=name)
        
        self.center = np.array(center,dtype=float).reshape(-1,1)
        if self.center.shape[0] == 3: # 3D sphere
            self.dimensionality = '3D'

        if not radius is None:
            radiusVector= np.zeros_like(self.center)
            radiusVector[0] = radius
            radiusPoint = self.center + radiusVector
        else:
            if radiusPoint is None:
                raise AttributeError('Either radius or radiusPoint is to be set.')
        radiusPoint = np.array(radiusPoint,dtype=float).reshape(-1,1)
        self.radius = np.linalg.norm(radiusPoint-self.center)
        self.radiusPoint = radiusPoint

        
        
        
    def plot(self,ax,transformation=None,**kwargs):# pragma: no cover
        if self.dimensionality == '2D':
            theta = np.linspace(0,np.pi*2,200)
            
            points = np.array([np.cos(theta),np.sin(theta)])*self.radius+self.center
            if not transformation is None:
                points = transformation(*points)
            ax.plot(points[0],points[1],**kwargs)
        elif self.dimensionality == '3D':
            
            u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
            x = np.cos(u)*np.sin(v)*self.radius+self.center[0]
            y = np.sin(u)*np.sin(v)*self.radius+self.center[1]
            z = np.cos(v)*self.radius+self.center[2]
            points = [x,y,z]
            if not transformation is None:
                points = transformation(*points)
            ax.plot_wireframe(*points, **kwargs)
    
    def call(self,x,y=None,z=None):
        if len(self.coordinates)>0:
            points = np.array([getattr(x,coord) for coord in self.coordinates])
        else:
            if self.dimensionality == '2D':
                if y is None:
                    if x.shape[0] != 2:
                        raise AttributeError('Dimensionality of x is to be (2,N), received {}'.format(x.shape))
                    points = x
                else:
                    points = np.array([x,y])
            elif self.dimensionality == '3D':
                if y is None or z is None:
                    if x.shape[0] != 3:
                        raise AttributeError('Dimensionality of x is to be (3,N), received {}'.format(x.shape))
                    points = x
                else:
                    points = np.array([x,y,z])
        
        pointShape = points.shape[1:]
        points.shape = (points.shape[0],-1)
        
        mask = np.linalg.norm(points-self.center,axis=0)<=self.radius
        mask.shape = pointShape
        
        if self.maskInside == False:
            mask = np.logical_not(mask)
        return mask
    
class indexMask(MaskingObject):
    dimensionality = '1D'
    
    def __init__(self,start,end,axis=0,maskInside=True,name=None):
        """Mask indices along a given axis.
        
        args:
            start (int): starting point included in mask
        
            end (int): end point not included in mask
            
        kwargs:
            axis (int): Axis along which masking is to be performed (default 0)
            
            maskInside (bool): If true, points inside is masked otherwise outside (default True)
        
        """
        super(indexMask,self).__init__(coordinates=None,maskInside=maskInside,name=name)
        self.axis = axis
        self.start = start
        self.end = end
        
    def plot(self,ax,transformation=None,*args,**kwargs):# pragma: no cover
        raise NotImplementedError('It is not possible to plot a 1D masks.')
    
    def call(self,x,*args):
        points = x#np.concatenate([[x],args])
        
        if not len(points.shape)>self.axis:
            raise AttributeError('Masking axis is {}, but shape of x is {}.'.format(self.axis,points))
        
        arange = np.arange(points.shape[self.axis]).flatten()
       
        transpose = np.arange(len(points.shape))
        transpose[self.axis] = 0
        transpose[0] = self.axis
        
        otherAxes = int(points.size/points.shape[self.axis])
        inside = np.array([np.logical_and(arange>=self.start,arange<self.end)]).flatten()
                
        # create mask with masking axis along 0
        newShape = np.array(points.shape)[transpose]
        mask = np.repeat(inside,otherAxes).reshape(newShape).transpose(transpose)
        
        
        if self.maskInside == False:
            mask = np.logical_not(mask)
        return mask


        
class CurratAxeMask(MaskingObject):
    dimensionality = '3D'
    
    def __init__(self,braggPeaks,dqx=None,dqy=None,dH=None,dK=None,dL=None,spurionType='both',maskInside=True,name=None):
        """Mask Currat-Axe spurions from provided Bragg peaks within wanted area.
        
        Args:
        
            - BraggPeaks (list): List of Bragg peaks to be used for mask. Shape is given by [[H1,K1,L1], [H2,K2,L2], ... [HN,KN,LN]]
            
        Kwargs:

            - dqx (float): Radius used for masking along qx (default None)

            - dqy (float): Radius used for masking along qy (default None)

            - dH (float): Radius used for masking along H (default None)

            - dK (float): Radius used for masking along K (default None)

            - dL (float): Radius used for masking along L (default None)

            - spurionType (str): Either monochromator, analyser or both (default 'both')
            
            - maskInside (bool): If true, points inside is masked otherwise outside (default True)
        
        """
        super(CurratAxeMask,self).__init__(coordinates=None,maskInside=maskInside,name=name)
        self.braggPeaks = braggPeaks
        self.dqx = dqx
        self.dqy = dqy
        self.dH = dH
        self.dK = dK
        self.dL = dL
        self.spurionType = spurionType
        
    def plot(self,ax,transformation=None,*args,**kwargs):# pragma: no cover
        raise NotImplementedError('It is not possible to plot a Currat-Axe masks currently.')
    
    def call(self,x,*args):
        if not isinstance(x,DataFile.DataFile):
            raise AttributeError('Expected mask to be called with '+str(type(DataFile.DataFile))+' object as argument. Received '+str(type(x)))
        
        
        return x.calculateCurratAxeMask(self.braggPeaks,dqx=self.dqx,dqy=self.dqy,dH=self.dH,dK=self.dK,dL=self.dL,spurionType=self.spurionType,maskInside=self.maskInside)


def parse(string,masks):
    for m in masks:
        locals()[m.name]=m
    
    return eval(string)



def flatten(t):
    return [item for sublist in t for item in sublist]


def extractSubMasks(m):
    if isinstance(m,(MultiMask)):
        if m.operation is np.logical_or:
            op = '+'
        else:
            op = '*'
        
        
        signs = ['-'*mask.negated for mask in m.masks]
        m.masks = [mask.bar() if mask.negated else mask for mask in m.masks]
        result = [extractSubMasks(mask) for mask in m.masks]
        eq = [res[0] for res in result]
        masks = flatten([res[1] for res in result])
    
        s = '('+op.join([s+str(e) for s,e in zip(signs,eq)])+')'

        maskSet = set()
        for mask in masks+m.masks:
            if not isinstance(mask,(MultiMask)):
                maskSet.add(mask)
        return s.format(*eq),maskSet
    else:
        return sympy.symbols(m.name),[m]
def extract(mask):
    eq,masks = extractSubMasks(mask)
    uniqueMasks = []
    for mask in masks:
        if not mask in uniqueMasks:
            uniqueMasks.append(mask)
    
    return str(sympy.expand(eq)),uniqueMasks


def load(file):
    with open(file,'rb') as f:
        mask = pickle.load(f)
    return mask


from MJOLNIR.Data import DataSet
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mayavi import mlab
from matplotlib.colors import ListedColormap
cmap = plt.cm.jet
#my_cmap = cmap(np.arange(cmap.N))
#my_cmap[:,-1] = np.linspace(0, 1, cmap.N)
#my_cmap = ListedColormap(my_cmap)

my_cmap = cmap

file = '/home/lass/Dropbox/PhD/Software/MJOLNIR/TestData/cameasim2018n000011.nxs'
DataObj = DataSet.DataSet(convertedFiles=file)

I = DataObj.I
qx = DataObj.qx
qy = DataObj.qy
energy = DataObj.energy
Norm = DataObj.Norm
Monitor = DataObj.Monitor


r = np.linalg.norm([qx,qy],axis=0)
theta = np.arctan2(qy,qx)

[I_bin,Monitor_bin,Normalization_bin,NormCount_bin],[r_bin,theta_bin,energy_bin] = \
DataSet.binData3D(0.01,np.deg2rad(0.5),0.5,[r.flatten(),theta.flatten(),energy.flatten()],data=I,norm=Norm,mon=Monitor)
Qx = np.cos(theta_bin)*r_bin
Qy = np.sin(theta_bin)*r_bin


Int = np.divide(I_bin*NormCount_bin,Monitor_bin*Normalization_bin)

# Plot energy slice of data
Eslice=1

if False:
    fig=plt.figure(figsize=(8,8))
    ax = plt.gca(projection='3d')
    for Eslice in range(len(energy_bin[0,0])-1):
    
        VMIN=1e-10
        VMAX=1e-6
        
        pc = ax.plot_surface(Qx[:,:,Eslice].T,Qy[:,:,Eslice].T,np.ones_like(Qy[:,:,Eslice]).T*energy_bin[0,0,Eslice],rstride=1,cstride=1,facecolors=my_cmap((Int[:,:,Eslice]/np.nanmax(Int[:,:,:])).T))#,vmin=VMIN,vmax=VMAX,zorder=10,cmap=my_cmap)
    #pc = ax.pcolormesh(Qx[:,:,Eslice].T,Qy[:,:,Eslice].T,Int[:,:,Eslice].T,vmin=VMIN,vmax=VMAX,zorder=10,cmap=my_cmap)
    ax.set_ylabel('$Q_y$ [1/AA]')
    ax.set_xlabel('$Q_x$ [1/AA]')
    plt.title('$\hbar \omega =$ {:.02f}'.format(np.mean(energy_bin[:,:,Eslice])) + ' meV')

else:
    for Eslice in range(len(energy_bin[0,0])-1):
        surf = mlab.mesh(Qx[:-1,:-1,Eslice].T,Qy[:-1,:-1,Eslice].T,energy_bin[0,0,Eslice]+Int[:,:,Eslice].T)#,warp_scale="auto")#np.ones_like(Qy[:,:,Eslice]).T*energy_bin[0,0,Eslice],color=(my_cmap((Int[:,:,Eslice].flatten()/np.nanmax(Int[:,:,:])).T)[:,:3]))
    
    xx = [np.min(Qx),np.max(Qx)]
    x0 = [xx[0],xx[0]]
    yy = [np.min(Qy),np.max(Qy)]
    y0 = [yy[0],yy[0]]
    zz = [np.min(energy_bin),np.max(energy_bin)]
    z0 = [zz[0],zz[0]]
    
    lensoffset=2.0
    mlab.plot3d(xx,y0,z0,line_width=0.01,tube_radius=0.01)# xaxis
    mlab.plot3d(x0,yy,z0,line_width=0.01,tube_radius=0.01)# yaxis
    mlab.plot3d(x0,y0,zz,line_width=0.01,tube_radius=0.01)# zaxis
    
    mlab.text3d(0.5*(np.sum(xx)), yy[0], zz[0], 'Qx [1/A]',scale=0.1)
    mlab.text3d(xx[0],0.5*(np.sum(yy)), zz[0], 'Qy [1/A]',scale=0.1)
    mlab.text3d(xx[0],yy[0],0.5*(np.sum(zz)), 'E [meV]',scale=0.1)
    
    #    mlab.plot3d(zx,zy+lensoffset,zz,line_width=0.01,tube_radius=0.01)
    #    mlab.plot3d(xx,xy+lensoffset,xz,line_width=0.01,tube_radius=0.01)
    
    #mlab.xlabel('Qx [1/A]')
    #mlab.ylabel('Qy [1/A]')
    #mlab.zlabel('E [meV]')
    mlab.show()


#plt.axis([-2.7, 2.7, -2.7, 2.7])
#ax.set_aspect('equal', 'datalim')
#plt.grid('on')
#plt.colorbar(pc)



# -*- coding: utf-8 -*-

import sys, os
sys.path.append('.')
sys.path.append('..')
sys.path.append('../..')


import numpy as np
import pickle as pickle
import matplotlib.pyplot as plt
from MJOLNIR import _tools
from MJOLNIR.Data import Mask

import functools
import warnings



# If an update has happened to the self.int, and the originalDataSet has a backgroundIntensities pointer array
def recalculateOnChange(func):
    @functools.wraps(func)
    def recalculateFunction(self,*args,**kwargs):
        if self.isChanged and hasattr(self.originalDataSet,'_backgroundIntensities'):
            #print('RECALCULATING!!')
            self.generateFullBackground()
            #print('Done',f'{self.isChanged=}', flush=True)
        #else:
        #    print('No recalculation needed')
        return func(self,*args,**kwargs)
    return recalculateFunction
        

class BackgroundObject(object):
    
    def __init__(self,originalDataSet,dQ=0.025, dE = 0.05, backgroundMask=None):
        """Object to stream line the background generation
        
        Args:
            
            - originalDataSet (DataSet): DataSet containing the data set used for the background
            
        Kwargs:
            
            - dQ (float): Step size along the q direction in the powder average bin in 1/AA (default 0.02)
            
            - dE (float): Step size along the energy direction in the powder average bin in meV (default 0.1)
            
            - backgroundMask (Mask): Masking object to utilize for background generation. If none use DataSet mask (default None)
        
        """
        
        self.originalDataSet = originalDataSet
        
        self.dQ = dQ
        self.dE = dE
        self.backgroundMask = backgroundMask
        
        # Flag to keep track of updates in model not yet applied
        self.isChanged = False
        
        
    def generatePowderBackground(self,dQ=None, dE = None):
        if not dQ is None:
            self.dQ = dQ
        if not dE is None:
            self.dE = dE
        
        ## Needed to reset data set after calculating background
        foregroundMask = self.originalDataSet.mask
        
        if self.backgroundMask is None:
            self.backgroundMask = foregroundMask
        else:
            if isinstance(self.backgroundMask,(Mask.MaskingObject)): # Is mask
                self.backgroundMask = self.backgroundMask(self.originalDataSet)
            
            self.originalDataSet.mask = self.backgroundMask
        # Find Q and E ranges
        
        qLength = np.linalg.norm([self.originalDataSet.qx.extractData(),self.originalDataSet.qy.extractData()],axis=0)
        
        QMax = np.nanmax(qLength) # QMin is always set to 0
        #print(QMax)
        Energies = self.originalDataSet.energy.extractData()
        
        EMin = np.nanmin(Energies)
        EMax = np.nanmax(Energies)#self.originalDataSet.energy.extractData().max()
                
        #print(EMin,EMax)
        I = self.originalDataSet.I.extractData()
        Monitor = self.originalDataSet.Monitor.extractData()
        Norm = self.originalDataSet.Norm.extractData()
        
        # Position in the powder average (QLength, Energy)
        positions2D = np.array([np.linalg.norm([self.originalDataSet.qx.extractData(),
                                                self.originalDataSet.qy.extractData()],axis=0),
                                self.originalDataSet.energy.extractData()])
    
        ## Reset mask 
        self.originalDataSet.mask = foregroundMask
        
        
        # Generate the bins with suitable extensions
        self.QBins = np.arange(0,QMax+self.dQ*1.1,self.dQ)
        self.EnergyBins = np.arange(EMin-self.dE*1.1,EMax+self.dE*1.1,self.dE)
    
        
        # Perform 2D histogram
        self._intensity,self._norm,self._monitor,self._counts = _tools.histogramdd(positions2D.T,bins=(self.QBins,self.EnergyBins),weights=[I,Norm,Monitor],returnCounts=True)
        
        
        
        ## Calcualte the intensities
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._int = np.divide(self.intensity*self.counts,self.monitor*self.norm)
        self.isChanged = True
        
    @recalculateOnChange
    def sample(self,positions,norm=None,monitor=None):#,returnAll=False):
        """Sample the background at the given positions (|Q|,E)
        
        Args:
            
            - positions (2D list): List of positions (|Q|, E) with shape N x 2
            
        Kwargs:
            
            - norm (list): List of normalizations of shape N (default None)
            
            - monitor (list): List of monitor of shape N (default None)
            
            - returnAll (bool): If True, returns intensity, monitor, norm, counts
            
            
        """
        BGBins = [self.QBins,self.EnergyBins]
            
        
        xQ = _tools.findFlattenedIndex(positions[0].reshape(-1,1), [BGBins[0]])-1
        yE = _tools.findFlattenedIndex(positions[1].reshape(-1,1), [BGBins[1]])-1

        inside = np.logical_and(xQ<len(self.QBins)-1,yE<len(self.EnergyBins)-1)
        newI = np.zeros_like(inside,dtype=float)
        
        
        
        newI[inside] = self.int[xQ[inside],yE[inside]]
        if not norm is None:
            newI[inside]*=norm.reshape(newI.shape)[inside]
        if not monitor is None:
            newI[inside]*=monitor.reshape(newI.shape)[inside]
        return newI
    
    def generateFullBackground(self):
        """Generate the full fetched background data set from the original one"""
        
        self.isChanged = False
        for df in self.originalDataSet:
            positions = np.asarray([np.linalg.norm([df.qx.flatten(),df.qy.flatten()],axis=0),df.energy.flatten()])
            df.backgroundIntensities = self.sample(positions,norm=df.Norm.flatten(),monitor=df.Monitor.flatten()).reshape(df.I.shape)
        
        files = self.originalDataSet.convertedFiles if len(self.originalDataSet.convertedFiles) > 0 else self.originalDataSet.dataFiles
        self.originalDataSet.backgroundIntensities = _tools.PointerArray('backgroundIntensities',files)
        
        
        
        
    @property
    def dQ(self):
        return self._dQ
    
    @dQ.getter
    def dQ(self):
        return self._dQ
    
    @dQ.setter
    def dQ(self,value):
        if value > 0:
            self._dQ = value
        else:
            raise AttributeError('Provided dQ is negative or null. Received {:}'.format(value))
    
    @property
    def dE(self):
        return self._dE
    
    @dE.getter
    def dE(self):
        return self._dE
    
    @dE.setter
    def dE(self,value):
        if value > 0:
            self._dE = value
        else:
            raise AttributeError('Provided dE is negative or null. Received {:}'.format(value))
            
    @property
    def int(self):
        if not hasattr(self,'_int'):
            warnings.warn('No background was found. Generating using standard parameters...')
            self.generatePowderBackground()
        return self._int
    
    
    
    @property
    def intensity(self):
        if not hasattr(self,'_intensity'):
            warnings.warn('No background was found. Generating using standard parameters...')
            self.generatePowderBackground()
        return self._intensity
    
    @property
    def monitor(self):
        if not hasattr(self,'_monitor'):
            warnings.warn('No background was found. Generating using standard parameters...')
            self.generatePowderBackground()
        return self._monitor
    
    @property
    def norm(self):
        if not hasattr(self,'_norm'):
            warnings.warn('No background was found. Generating using standard parameters...')
            self.generatePowderBackground()
        return self._norm
    
    @property
    def counts(self):
        if not hasattr(self,'_counts'):
            warnings.warn('No background was found. Generating using standard parameters...')
            self.generatePowderBackground()
        return self._counts
    
    
    
    
    def plotPowderAverage(self,ax=None):
        if ax is None:
            fig,ax = plt.subplots()
            
        if not hasattr(self,'Int'):
            self.generatePowderBackground()
            
        
        X,Y = np.meshgrid(self.QBins,self.EnergyBins)
        ax.p = ax.pcolormesh(X,Y,self.int.T)
        ax.set_xlabel('|Q| [1/A]')
        ax.set_ylabel('E [meV]')
        ax.get_figure().colorbar(ax.p)
        
        def set_clim(ax,vmin,vmax):
            ax.p.set_clim(vmin,vmax)
            
        ax.set_clim = lambda vmin,vmax: set_clim(ax,vmin,vmax)
        
        return ax


import numpy as np
from MJOLNIR._tools import *

import os

def test_minMax():
    L = np.random.rand(10,3,2)
    minmax = minMax(L)
    
    assert(len(minmax)==2)
    assert(np.isclose(minmax[0],np.min(L)))
    assert(np.isclose(minmax[1],np.max(L)))

    minmax = np.array(minMax(L,axis=-1))
    assert(minmax.shape==(2,10,3))

    minmax = np.array(minMax(L,axis=0))
    assert(minmax.shape==(2,3,2))

def test_unitVector():
    V = np.array([1.0,2.0,3.0])
    v = unitVector(V)
    assert(np.isclose(np.linalg.norm(v),1.0))


def test_rotations():
    vectors = [np.array([0,0,3.0]),np.array([1.0,0.0,0.0]),np.array([0.0,1.0,0.0]),np.random.rand(3)]
    rotations = [rotate2X(v) for v in vectors]
    rotVector = [np.dot(rotations[i],vectors[i]) for i in range(len(vectors))]
    for i in range(len(rotVector)):
        assert(np.isclose(np.linalg.norm(vectors[i]),np.linalg.norm(rotVector[i])))
        print(rotVector[i][0],np.linalg.norm(rotVector[i]))
        assert(np.isclose(rotVector[i][0],np.linalg.norm(rotVector[i])))


def test_vectorAngle():
    v1 = np.array([1,0,0])
    v2 = np.array([0,1,0])
    v3 = np.array([1,1,1])
    theta1 = vectorAngle(v1,v2)
    theta2 = vectorAngle(v1,v3)
    theta3 = vectorAngle(v2,v3)
    print(theta2)
    assert(np.isclose(theta1,np.pi/2))
    assert(np.isclose(theta2,theta3))
    assert(np.isclose(theta2,0.955316618125))
    
def test_Rotation_matrix():
    M1 = rotationMatrix(0,0,0)

    assert(np.all(np.isclose(M1,np.identity(3))))

    M2 = rotationMatrix(90,0,0)
    M3 = rotationMatrix(np.pi/2,np.pi/3,np.pi/4,format='rad')
    M4 = rotationMatrix(90,60,45)
    
    assert(np.all(np.isclose(M2,np.array([[1,0,0],[0,0,-1],[0,1,0]]))))
    assert(np.all(np.isclose(M3,M4)))
    
    M4_check = np.array([[3.53553391e-01,6.12372436e-01,7.07106781e-01],[3.53553391e-01,6.12372436e-01,-7.07106781e-01],[-8.66025404e-01,5.00000000e-01,3.06161700e-17]])
    assert(np.all(np.isclose(M4,M4_check)))




def test_binEdges():
    values = np.exp(np.linspace(-0.1,1,101)) # Generate non-linear points
    #values[0]-tolerance/2.0 and ends at values[-1]+tolerance/2.0.
    minBin = 0.1
    bins = binEdges(values,minBin)

    assert(np.isclose(bins[0],values[0]-0.1*minBin)) # First bin starts at values[0]-tolerance*0.1
    assert(bins[-1]>=values[-1]+0.1*minBin) # Last bin ends at values[-1]+tolerance*0.1
    assert(np.all(np.diff(bins)>=minBin*0.99)) # Assert that all bins are at least of size minBin

    binsCut = binEdges(values,0.01,startPoint=0.0)
    assert(binsCut[0]>=0.0)

    binsCut = binEdges(values+np.min(values)+0.0099,0.01,startPoint=0.0,endPoint=0.5)
    assert(binsCut[0]>=0.0)
    assert(binsCut[-1]<=0.5)

    binsCut = binEdges(values+np.min(values)+0.0101,0.01,startPoint=0.0,endPoint=0.5)
    assert(binsCut[0]>0.01)

    binsCut = binEdges(np.linspace(0,0.95,101),0.01,startPoint=0.0,endPoint=1.0101)
    assert(binsCut[-1]<=1.0)

    binsCut = binEdges(values-np.max(values)+1.0,0.01,startPoint=0.0,endPoint=1.01)
    assert(binsCut[-1]<=1.01)

    
def test_fileListGenerator():
    numberStr = '0,20-23-24,4000'
    year = 2018
    folder = os.path.join('','home','camea')
    try:
        files = fileListGenerator(numberStr,folder,year)
        assert False # Too many dasches
    except AttributeError:
        assert True

    numberStr = '0,20-23,4000'

    files = fileListGenerator(numberStr,folder,year)
    filesCorrect = np.array([
        os.path.join('','home','camea','camea2018n000000.hdf'), os.path.join('','home','camea','camea2018n000020.hdf'), 
        os.path.join('','home','camea','camea2018n000021.hdf'), os.path.join('','home','camea','camea2018n000022.hdf'), 
        os.path.join('','home','camea','camea2018n000023.hdf'), os.path.join('','home','camea','camea2018n004000.hdf')])
    assert(np.all(filesCorrect == np.array(files)))

    # opposite direction
    numberStr = '23-20'
    filesCorrect = np.array([
        os.path.join('','home','camea','camea2018n000023.hdf'),os.path.join('','home','camea','camea2018n000022.hdf'), 
        os.path.join('','home','camea','camea2018n000021.hdf'),os.path.join('','home','camea','camea2018n000020.hdf')])

    files = fileListGenerator(numberStr,folder,year)
    assert(np.all(filesCorrect == np.array(files)))

def test_RoundBinning():
    
    X = np.random.rand(2,3000)
    binning = [0.05,0.1,0.2]
    try:
        RoundBinning(X,binning)
    except AttributeError:
        assert True
    
    binning = [0.05,0.05]
    BP, I, C  = RoundBinning(X,binning)

    Int = np.random.rand(3000)
    BP2,I2,C2,Data = RoundBinning(X,binning[0],Int)

    assert(np.all(BP==BP2))
    assert(np.all(I==I2))
    assert(np.all(C==C2))
    assert(np.all(Data.shape[0]==BP.shape[1]))


def test_generateLabel():
    v = [1,-1,2]
    label = generateLabel(v)
    assert(label=='(H, -K, 2L)')

    v = np.array([-0.5,22.0000,-3.0])
    label = generateLabel(v)
    assert(label=='(-0.5H, 22K, -3L)')


def test_generateLabelDirection():
    v = [1,-1,2]
    label = generateLabelDirection(v)
    assert(label=='(H, -H, 2H)')

    v = np.array([0.0,22.0000,-3.0])
    label = generateLabelDirection(v)
    assert(label=='(0, 22K, -3K)')

    v = np.array([0.0,0.0000,-3.333])
    label = generateLabelDirection(v)
    assert(label=='(0, 0, -3.333L)')

def test_molarMassCalculation():
    water = 18.01528 # g/mol
    value,elements = calculateMolarMass('H2O',returnElements=True)
    assert(np.isclose(value,water,atol=1e-4))
    assert(elements=={'H':2,'O':1})

    glucose = 180.156 # g/mol
    assert(np.isclose(calculateMolarMass('C6H12O6'),glucose,atol=1e-4))

    YMnO3 = 191.84 # g/mol
    assert(np.isclose(calculateMolarMass('YMnO3'),YMnO3,atol=1e-4))

    YBCO = 666.19 # g/mol
    value = calculateMolarMass('YBa2Cu3O7')
    assert(np.isclose(value,YBCO,atol=1e-4))
    
    
    Crazy = 'U2.3(H3O2.2)2.2Be((He2.2Ge)2Fe)3.3'
    weight = 2724.6618342799998
    elem = {'H': 13.2,  'O': 9.68,  'He': 29.04,  'Ge': 13.2,  'Fe': 6.6,  'U': 4.6,  'Be': 2.0}
    value,elements = calculateMolarMass(Crazy,formulaUnitsPerUnitCell=2,returnElements=True)
    assert(np.isclose(value,weight,atol=1e-4))
    assert(np.all([np.isclose(elem[key],elements[key],atol=1e-8) for key in elem.keys()]))

    
def test_absoluteNormalization():
    sampleMass = 6.2
    normFactor = calculateAbsoluteNormalization(sampleChemicalFormula='MnF2',formulaUnitsPerUnitCell=2,sampleMass=sampleMass,correctVanadium=True)
    # Known value for MnF2
    assert(np.isclose(normFactor,2.0016916920401816e-07))

    vanadiumMass=15.25
    vanadiumMonitor=1e5
    vanadiumSigmaIncoherent = 5.08
    constants = 4*np.pi/13.77
    Van = calculateAbsoluteNormalization(sampleChemicalFormula='V',formulaUnitsPerUnitCell=1,sampleMass=vanadiumMass,correctVanadium=True)
    

    Van*=vanadiumMonitor*vanadiumSigmaIncoherent/constants
    assert(np.isclose(Van,1.0))

    # Same as above but use the molecular mass in stead of calculated
    sampleMass = 6.2
    normFactor2 = calculateAbsoluteNormalization(sampleMolarMass=92.9349*2,sampleMass=sampleMass,correctVanadium=True)
    # Known value calcualted above
    assert(np.isclose(normFactor,normFactor2))



def test_KWavelength():
    k = np.pi
    wavelength = WavelengthK(k)
    assert(np.isclose(wavelength,2.0))
    
    k_prime = KWavelength(wavelength)
    assert(np.isclose(k,k_prime))
    
def test_KWavelength_multidim():
    k = np.random.rand(10,20)*2*np.pi
    wavelength = WavelengthK(k)
    
    k_prime = KWavelength(wavelength)
    assert(np.all(np.isclose(k,k_prime)))
    
    
def test_EnergyK():
    E = 5.0
    k = KEnergy(E)
    l = WavelengthEnergy(E)
    assert(np.isclose(l,4.044851376460777))
    
    E_prime = EnergyK(k)
    assert(np.isclose(E,E_prime))
    
def test_WnergyK_multidim():
    E = np.random.rand(10,20)*2*np.pi
    wavelength = WavelengthEnergy(E)
    
    E_prime = EnergyWavelength(wavelength)
    assert(np.all(np.isclose(E,E_prime)))
    
    
def test_ScatteringAngle():
    d = 3.355 # AA
    E = 5.00 # meV
    K = KEnergy(E)
    W = WavelengthEnergy(E)
    
    A4E = ScatteringAngle(d,Energy=E)
    A4K = ScatteringAngle(d,K=K)
    A4W = ScatteringAngle(d,Wavelength=W)
    A4ERad = ScatteringAngle(d,Energy=E,degrees=False)
    A4ERad = np.rad2deg(A4ERad)
    
    A4True = 74.14275085067898
    
    assert(np.all(np.isclose([A4E,A4K,A4W,A4ERad],A4True)))
    
def test_DSpacing():
    twoTheta = 74.14275085067898#
    E = 5.00 # meV
    K = KEnergy(E)
    W = WavelengthEnergy(E)
    
    dE = DSpacing(twoTheta,Energy=E)
    dK = DSpacing(twoTheta,K=K)
    dW = DSpacing(twoTheta,Wavelength=W)
    dERad = DSpacing(np.deg2rad(twoTheta),Energy=E,degrees=False)
    
    
    dTrue = 3.355 # AA
    
    assert(np.all(np.isclose([dE,dK,dW,dERad],dTrue)))
    
def test_ScatteringAngle_Errors():
    d = 3.355 # AA
    E = 5.00 # meV
    K = KEnergy(E)
    W = WavelengthEnergy(E)
    
    try: 
        _ = ScatteringAngle(d,Energy=E,K=K)
        assert False
    except AttributeError:
        assert True
    
    try: 
        _ = ScatteringAngle(d,Energy=E,Wavelength=W)
        assert False
    except AttributeError:
        assert True
    
    try: 
        _ = ScatteringAngle(d,Energy=E,Wavelength=W,K=K)
        assert False
    except AttributeError:
        assert True
    
    try: 
        _ = ScatteringAngle(d)
        assert False
    except AttributeError:
        assert True
        
        
def test_DSpacing_Errors():
    twoTheta = 3.355 # AA
    E = 5.00 # meV
    K = KEnergy(E)
    W = WavelengthEnergy(E)
    
    try: 
        _ = DSpacing(twoTheta,Energy=E,K=K)
        assert False
    except AttributeError:
        assert True
    
    try: 
        _ = DSpacing(twoTheta,Energy=E,Wavelength=W)
        assert False
    except AttributeError:
        assert True
    
    try: 
        _ = DSpacing(twoTheta,Energy=E,Wavelength=W,K=K)
        assert False
    except AttributeError:
        assert True
    
    try: 
        _ = DSpacing(twoTheta)
        assert False
    except AttributeError:
        assert True
    

def test_scaling():# Test standard scaling functions used for rescaling of energy bins

    x = np.random.rand(20)*20-10 # points between -10 and 10
    assert(np.all(np.isclose(x,rescale(scale(x)))))
    assert(np.all(np.isclose(x,rescale(scale(x,f=0.1),f=0.1))))

def test_getNext():
    
    data = [1,2,3,4,5,6,7,8]

    assert(len(data)==8)
        
        
    for d in getNext(data,delete=True):
        print(d)
    assert(len(data)==0)
    
    data = [1,2,3,4,5,6,7,8]
    
    for d in getNext(data,delete=False):
        print(d)
    
    assert(len(data)==8)
    


import sys, os
sys.path.append('.')
sys.path.append('..')
sys.path.append('../..')
import scipy
import matplotlib.pyplot as plt
import numpy as np
import h5py as hdf
import warnings
from MJOLNIR import _tools
import datetime
import math
from MJOLNIR import TasUBlibDEG as TasUBlib
from MJOLNIR._tools import Marray
import MJOLNIR.Data.DataFile


def cosd(x):
    return np.cos(np.deg2rad(x))

def sind(x):
    return np.sin(np.deg2rad(x))

def camelCase(string,split='_'):
    """Convert string to camel case from <split> seperated"""

    if not split in string:
        return string
    splitString = string.split(split)
    first = splitString[0]
    others = [x.title() for x in splitString[1:]]
    
    combi = [first]+others
    return ''.join([str(x) for x in combi])

class Sample(object):
    """Sample object to store all information of the sample from the experiment"""
    @_tools.KwargChecker()
    def __init__(self,a=2.0*np.pi,b=2.0*np.pi,c=2.0*np.pi,alpha=90,beta=90,gamma=90,sample=None,name='Unknown',projectionVector1=None, projectionVector2 = None,recalculateUB=False):
        if isinstance(sample,hdf._hl.group.Group):
            self.name = str(np.array(sample.get('name'))[0].decode())

            self.planeNormal = np.array(sample.get('plane_normal'))
            
            self.polarAngle = np.array(sample.get('polar_angle'))
            self.rotationAngle = np.array(sample.get('rotation_angle'))
            self.unitCell = np.array(sample.get('unit_cell'))
            self.plane_vector1 = np.array(sample.get('plane_vector_1'))
            self.plane_vector2 = np.array(sample.get('plane_vector_2'))
            crossProduct = np.cross(self.plane_vector1[:3],self.plane_vector2[:3])
            if recalculateUB is True: # Recalculate the ub from the given peaks (ignoring sgu and sgl!)
                self.plane_vector1 = np.delete(self.plane_vector1,3)
                self.plane_vector1[5:7] = 0.0
                self.plane_vector2 = np.delete(self.plane_vector2,3)
                self.plane_vector2[5:7] = 0.0
                self.orientationMatrix = TasUBlib.calcTasUBFromTwoReflections(self.cell, self.plane_vector1, self.plane_vector2)
            elif recalculateUB == 0 and type(recalculateUB) == bool:
                self.orientationMatrix = np.array(sample.get('orientation_matrix'))*2*np.pi
            else:
                self.orientationMatrix = np.array(sample.get('orientation_matrix'))
            if not np.all(np.isclose(crossProduct,[0,0,0])):
                self.planeNormal = crossProduct
            self.A3Off = np.array([0.0])#
            if not np.isclose(np.linalg.norm(self.plane_vector1[:3].astype(float)),0.0) or not np.isclose(np.linalg.norm(self.plane_vector2[:3].astype(float)),0.0): # If vectors are not zero
                self.projectionVector1,self.projectionVector2 = calcProjectionVectors(self.plane_vector1.astype(float),self.plane_vector2.astype(float))
            else:
                self.projectionVector1,self.projectionVector2 = [np.array([1.0,0.0,0.0]),np.array([0.0,1.0,0.0])]
            self.initialize()
            self.calculateProjections()

            attributes = ['azimuthal_angle','x','y','sgu','sgu_zero','sgl','sgl_zero']
            values = [camelCase(x) for x in attributes]
            for att, val in zip(attributes,values):
                setattr(self,val,np.array(sample.get(att)))
            
            
        elif np.all([a is not None,b is not None, c is not None]):
            self.unitCell = np.array([a,b,c,alpha,beta,gamma])
           
            self.polarAngle = np.array(None)
            self.rotationAngle = np.array(0)
            self.name=name
            if projectionVector1 is None or projectionVector2 is None:
                projectionVector1,projectionVector2 = [np.array([1.0,0.0,0.0,0.0,-37.553,0.0,0.0,5.0,5.0]),np.array([0.0,1.0,0.0,90.0,-37.553,0.0,0.0,5.0,5.0])]

            r1 = projectionVector1
            r2 = projectionVector2
            self.plane_vector1 = r1
            self.plane_vector2 = r2

            self.planeNormal = np.cross(self.plane_vector1[:3],self.plane_vector2[:3])
            
            
            cell = TasUBlib.calcCell(self.unitCell)

            self.orientationMatrix = TasUBlib.calcTasUBFromTwoReflections(cell, r1, r2)#(np.pi*2)
            #self.orientationMatrix = TasUBlib.calcTasUBFromTwoReflections(self.cell,self.plane_vector1,self.plane_vector2)
            self.projectionVector1,self.projectionVector2 = calcProjectionVectors(self.plane_vector1.astype(float),self.plane_vector2.astype(float))#,self.planeNormal.astype(float))
            self.initialize()
            self.calculateProjections()
        else:
            print(sample)
            print(a,b,c,alpha,beta,gamma)
            raise AttributeError('Sample not understood')

            
    @property
    def unitCell(self):
        return self._unitCelll

    @unitCell.getter
    def unitCell(self):
        return np.array([self.a,self.b,self.c,self.alpha,self.beta,self.gamma])#self._unitCell

    @unitCell.setter
    def unitCell(self,unitCell):
        self._unitCell = unitCell
        self.a = unitCell[0]
        self.b = unitCell[1]
        self.c = unitCell[2]
        self.alpha = unitCell[3]
        self.beta  = unitCell[4]
        self.gamma = unitCell[5]
        self.updateCell()
        
    @property
    def a(self):
        return self._a

    @a.getter
    def a(self):
        return self._a

    @a.setter
    def a(self,a):
        if a>0:
            self._a = a
        else:
            raise AttributeError('Negative or null given for lattice parameter a')

    @property
    def b(self):
        return self._b

    @b.getter
    def b(self):
        return self._b

    @b.setter
    def b(self,b):
        if b>0:
            self._b = b
        else:
            raise AttributeError('Negative or null given for lattice parameter b')

    @property
    def c(self):
        return self._c

    @c.getter
    def c(self):
        return self._c

    @c.setter
    def c(self,c):
        if c>0:
            self._c = c
        else:
            raise AttributeError('Negative or null given for lattice parameter c')


    @property
    def alpha(self):
        return self._alpha

    @alpha.getter
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self,alpha):
        if alpha>0 and alpha<180:
            self._alpha = alpha
        else:
            raise AttributeError('Negative,null or above 180 degrees given for lattice parameter alpha')

    @property
    def beta(self):
        return self._beta

    @beta.getter
    def beta(self):
        return self._beta

    @beta.setter
    def beta(self,beta):
        if beta>0 and beta<180:
            self._beta = beta
        else:
            raise AttributeError('Negative,null or above 180 degrees given for lattice parameter beta')

    @property
    def gamma(self):
        return self._gamma

    @gamma.getter
    def gamma(self):
        return self._gamma

    @gamma.setter
    def gamma(self,gamma):
        if gamma>0 and gamma<180:
            self._gamma = gamma
        else:
            raise AttributeError('Negative,null or above 180 degrees given for lattice parameter gamma')


    def __eq__(self,other):
        if not isinstance(other,type(self)):
            return False
        return np.all(np.isclose(self.unitCell,other.unitCell))#,\
        #np.all(self.orientationMatrix==other.orientationMatrix)])

    def initialize(self):
        """Initialize the Sample object. Automatically called during __init__method."""
        # From http://gisaxs.com/index.php/Unit_cell
        self.updateCell()
        self.B = TasUBlib.calculateBMatrix(self.cell)
        #self.reciprocalMatrix = np.array([self.reciprocalVectorA,self.reciprocalVectorB,self.reciprocalVectorC]).T

    def calculateProjections(self):
        """Calculate projections and generate projection angles."""
        checks = np.array(['unitCell','orientationMatrix','projectionVector1','projectionVector2']) # 
        boolcheck = np.logical_not(np.array([hasattr(self,x) for x in checks]))
        if np.any(boolcheck):
            raise AttributeError('Sample object is missing: {}.'.format(', '.join(str(x) for x in checks[boolcheck])))
        
        if self.projectionVector1[np.argmax(np.abs(self.projectionVector1))]<0:
            self.projectionVector1*=-1
        if self.projectionVector2[np.argmax(np.abs(self.projectionVector2))]<0:
            self.projectionVector2*=-1

        V1 = self.projectionVector1.copy()
        #V1/=np.linalg.norm(V1)
        V2 = self.projectionVector2.copy()
        #V2/=np.linalg.norm(V2)
        pV1Q = np.dot(self.B,V1)
        pV2Q = np.dot(self.B,V2)
        self.projectionAngle = _tools.vectorAngle(pV1Q,pV2Q)

        if np.isclose(0.0,self.projectionAngle):
            raise AttributeError("The provided orientations are equal.")

        
        UB= self.orientationMatrix
        self.UB = UB

        self.orientationMatrixINV = np.linalg.inv(UB)
        

        p23 = np.array([[1,0,0],[0,1,0]]) # To extract Qx, Qy only
        PM = np.array([V1,V2]).T # Projection matrix
        self.PM = PM
        self.convert = np.dot(p23,np.einsum('ij,jk->ik',UB,PM)) # Convert from projX,projY to Qx, Qy
        self.convertHKL = np.dot(p23,UB) # Convert from HKL to Qx, Qy

        # Calculate 'misalignment' of the projection vector 1
        try:
            self.theta = -TasUBlib.calcTasMisalignment(UB,self.planeNormal,V1)
        except AttributeError:
            self.theta = 0
        
        self.RotMat = _tools.Rot(self.theta) # Create 2x2 rotation matrix
        
        self.convertinv = np.linalg.inv(self.convert) # Convert from Qx, Qy to projX, projY

        self.convertHKLINV = _tools.invert(self.convertHKL) # Convert from Qx, Qy to HKL

        # Calcualte RotationMatrix for UB as block diagonal matrix. Only Qx,Qy part rotates as'
        # calculated in RotMat
        self.RotMat3D = np.eye(3)
        self.RotMat3D[:2,:2] = self.RotMat
        #self.orientationMatrixINV = np.linalg.inv(np.dot(self.RotMat3D,UB))
        self.orientationMatrixINV = np.linalg.inv(self.UB)
        

    def tr(self,p0,p1):
        """Convert from projX, projY coordinate to Qx,QY coordinate."""
        p0, p1 = np.asarray(p0), np.asarray(p1)
        
        P = np.array([p0,p1])
        
        Pos = np.einsum('ij,j...->i...',self.convertinv,P)
        return Pos[0],Pos[1]
        
    def inv_tr(self, x,y):
        """Convert from Qx,QY  coordinate to projX, projY coordinate."""
        x, y = np.asarray(x), np.asarray(y)
        P = np.array([x,y])
        Pos = np.einsum('ij,j...->i...',self.convert,P)
        return Pos[0],Pos[1]   


    def format_coord(self,x,y):
        """Format coordinates from QxQy in rotated frame into HKL."""
        x, y = np.asarray(x), np.asarray(y)
        rlu = self.calculateQxQyToHKL(x,y)#np.dot(self.orientationMatrixINV,np.array([x,y,0]))
        return "h = {0:.3f}, k = {1:.3f}, l = {2:.3f}".format(rlu[0],rlu[1],rlu[2])

    def calculateQxQyToHKL(self,x,y):
        """convert from Qx,Qy to HKL."""
        pos = np.array([x,y,np.zeros_like(x)])
        return np.einsum('ij,j...->i...',self.orientationMatrixINV,pos)

    def calculateHKLToQxQy(self,H,K,L):
        """convert HKL to Qx,Qy."""
        pos = np.array([H,K,L])
        return np.einsum('ij,j...->i...',self.orientationMatrix,pos)[:2]

    def calculateHKLtoProjection(self,H,K,L):
        """convert from projections to HKL."""
        HKL = np.array([H,K,L])
        #points = np.einsum('i...,ij...->i...',HKL,self.PM)
        points = np.einsum('ij,j...->i...',_tools.invert(self.PM),HKL)
        return points


    def __str__(self):
        returnStr = 'Sample ' + self.name + '\n'
        #if not self.temperature is None: returnStr+= 'Temperatur: '+str(self.temperature)+'\n'
        #if not self.magneticField is None: returnStr+= 'Magnetic Field: '+str(self.magneticField)+'\n'
        #if not self.electricField is None: returnStr+= 'Electric Field: '+str(self.electricField)+'\n'
        returnStr+= 'Unit cell: \n' + str(self.unitCell) + '\n'
        returnStr+= 'Orientation matrix: \n' + str(self.orientationMatrix) +'\n'

        return returnStr

    @_tools.KwargChecker()
    def CurratAxe(self,Ei,Ef,Bragg,spurionType='Monochromator',HKL=False,Projection=False):
        """Function to calculate Currat-Axe position in QxQy coordinate system.
    
        Args:
            
            - Ei (float/list): Incoming energy in meV.
            
            - Ef (float/list): Outgoing energy in meV.
            
            - Bragg (list): Bragg peak in HKL or list of.
            
        Kwargs:
            
            - spurionType (str): Either "Monochromator" or "Analyser" for origin of "wrong" energy (default "Monochromator").
            
            - HKL (bool): Whether or not to recalculate to HKL instead of Qx, Qy, Qz (default False).
            
            - Projection (Bool): Whether or not to recalculate to Projection vectors instead of Qx, Qy, Qz (default False).
            
        Returns:
            
            - Position (list): List of size (len(Bragg),len(Ei),len(Ef),3), where last axis is Qx, Qy, Qz
            
        """
        A3Off = self.theta
        
        Ei = np.asarray(Ei).flatten() # Shape (m)
        Ef = np.asarray(Ef).flatten() # Shape (n)
        Bragg = np.asarray(Bragg).reshape(-1,3) # shape (l,3)
        
        QlLocal = []
        if spurionType.lower() == 'monochromator':
            for B in Bragg:
                Ql = []
                Angles = np.array([TasUBlib.calcTasQAngles(self.orientationMatrix,self.planeNormal,1.0,0.0,np.array([B[0],B[1],B[2],e,e]))[:2] for e in Ef])
                for ei in Ei:
                    Ql.append(np.array([TasUBlib.calcTasQH(self.orientationMatrixINV,angle,ei,e,0) for angle,e in zip(Angles,Ef)])[:,1])
                QlLocal.append(Ql)
        elif spurionType.lower() == 'analyser':
            for B in Bragg:
                Ql = []
                for ei in Ei:
                    Angles2 = np.array(TasUBlib.calcTasQAngles(self.orientationMatrix,self.planeNormal,1.0,0.0,np.array([B[0],B[1],B[2],ei,ei]))[:2])
                    Ql.append(np.array([TasUBlib.calcTasQH(self.orientationMatrixINV,Angles2,ei,e,0.0) for e in Ef])[:,1]) # Extract Qx,Qy
                QlLocal.append(Ql)
        else:
            raise AttributeError('Provided spurionType not understood. Expected "Monochromator" or "Analyser" but received "{}".'.format(spurionType))

        returnVal = np.array(QlLocal) # Shape (l,m,n,3)
        
        if HKL == True or Projection == True: # need to calculate HKL for Projection calculation
            returnValShape = np.array(returnVal.shape)
            returnVal = self.calculateQxQyToHKL(returnVal[:,:,:,0].flatten(),returnVal[:,:,:,1].flatten())
            if Projection == True:
                
                toProjection = self.calculateHKLtoProjection(returnVal[0],returnVal[1],returnVal[2])
                returnVal = np.array(self.inv_tr(toProjection[0],toProjection[1]))
                returnValShape[-1]=2 # reshape Qx,Qy,Qz dimension to P1,P2 (3 -> 2)
            
            returnVal.shape = returnValShape # Shape (l,m,n,3) or (l,m,n,2)
        return returnVal

    def updateCell(self,unitCell=None):
        """Update cell parameters with current unit cell values.

        Kwargs:
            
            - unitCell (list): List of a,b,c,alpha,beta,gamma. If None, use self.unitCell (default None)

        """
        if unitCell is None:
            unitCell = self.unitCell
        else:
            self.unitCell = unitCell

        self.realVectorA = np.array([self.a,0,0])
        self.realVectorB = self.b*np.array([cosd(self.gamma),sind(self.gamma),0.0])#np.dot(np.array([self.b,0,0]),rotationMatrix(0,0,self.gamma))
        self.realVectorC = self.c*np.array([cosd(self.beta),(cosd(self.alpha)-cosd(self.beta)*cosd(self.gamma))/sind(self.gamma),
        np.sqrt(1-cosd(self.beta)**2-((cosd(self.alpha)-cosd(self.beta)*cosd(self.gamma))/sind(self.gamma))**2)])#np.dot(np.array([self.c,0,0]),rotationMatrix(0,self.beta,0))
        
        self.volume = np.abs(np.dot(self.realVectorA,np.cross(self.realVectorB,self.realVectorC)))
        self.reciprocalVectorA = 2*np.pi*np.cross(self.realVectorB,self.realVectorC)/self.volume
        self.reciprocalVectorB = 2*np.pi*np.cross(self.realVectorC,self.realVectorA)/self.volume
        self.reciprocalVectorC = 2*np.pi*np.cross(self.realVectorA,self.realVectorB)/self.volume
        

        bv1,bv2,bv3 = self.reciprocalVectorA,self.reciprocalVectorB,self.reciprocalVectorC
        a1,a2,a3,alpha1,alpha2,alpha3= self.unitCell
        
        b1,b2,b3 = [np.linalg.norm(x) for x in [bv1,bv2,bv3]]
        beta1 = np.rad2deg(_tools.vectorAngle(bv2,bv3))
        beta2 = np.rad2deg(_tools.vectorAngle(bv3,bv1))
        beta3 = np.rad2deg(_tools.vectorAngle(bv1,bv2))
        self.cell = [a1,a2,a3,b1,b2,b3,alpha1,alpha2,alpha3,beta1,beta2,beta3]


    def updateSampleParameters(self,unitCell):
        """Update the sample parameters and change UB matrix as well.

        Args: 

            - unitCell (list): List of cell parameters. If None use self.unitCell

        """
        
        UB = self.UB
        B = self.B
        U = np.dot(UB,np.linalg.inv(B))
        
        self.updateCell(unitCell)
        newB = TasUBlib.calculateBMatrix(self.cell)
        newUB = np.dot(U,newB)
        
        self.UB = newUB
        self.orientationMatrix = newUB
        self.calculateProjections()




def calcProjectionVectors(R1,R2,norm=None):
    r1 = R1[:3]
    r2 = R2[:3]
    if not norm is None:
        NV = norm
    else:
        NV = np.cross(r1,r2)
    NV/= np.linalg.norm(NV)
    
    Zeros = np.isclose(NV,0.0)
    if np.sum(Zeros)==3:
        raise AttributeError('The two plane vectors are equivalent, {}, {}!'.format(r1,r2))
    
    
    if np.sum(Zeros) == 2 or np.sum(Zeros)==1: # Easy case where the two vectors are to be along the x, y, or z directions
        if Zeros[0] == True:
            V1 = np.array([1.0,0.0,0.0])
            V2 = np.cross(NV,V1)
        elif Zeros[1]:
            V1 = np.array([0.0,1.0,0.0])
            V2 = np.cross(NV,V1)
        else:
            V1 = np.array([0.0,0.0,1.0])
            V2 = np.cross(NV,V1)
    else: # The tricky case of all vectors having non-zero components.
        V1 = r1
        V2 = r2
            
    V1 = _tools.LengthOrder(V1)
    V2 = _tools.LengthOrder(V2)
    #for V in [V1,V2]: # Flip sign if needed
    #    maxArg = np.argmax(np.abs(V))
    #    V*=np.sign(V[maxArg])
    return V1,V2


def calculateSample(cell,HKL1,HKL2,A3R1,A3R2,Ei,Ef):
    """
    Generate a sample from two reflections, their A3 positions, and incoming and outgoing energies
    
        - HKL1 ([HKL]): First alignment point

        - HKL2 ([HKL]): Second alignment point
        
        - A3R1 (float): A3 value of reference peak 1 (deg)
        
        - A3R2 (float): A3 value of reference peak 2 (deg)
        
        - Ei (float): Incoming energy used (meV)
        
        - Ef (float): Outgoing energy used (meV)
        
    """
    
    if np.isclose(A3R1,A3R2):
        raise AttributeError('A3 positions of the two peaks are identical.')
    
    if np.all(np.isclose(HKL1,HKL2)):
        raise AttributeError('Provided HKL positions are identical.')
    
    B = TasUBlib.calculateBMatrix(TasUBlib.calcCell(cell))
    
    A4R1 = TasUBlib.calTwoTheta(B, [*HKL1,Ei,Ef], -1)
    A4R2 = TasUBlib.calTwoTheta(B, [*HKL2,Ei,Ef], -1)
    
    r1 = np.array([*HKL1,A3R1, A4R1, 0.0,0.0, Ei,Ei]) # H K L A3 A4 SGU SGL Ei Ef
    r2 = np.array([*HKL2,A3R2, A4R2, 0.0,0.0, Ei,Ei]) # H K L A3 A4 SGU SGL Ei Ef
    
    return Sample(*cell,projectionVector1=r1,projectionVector2=r2)





import numpy as np
from MJOLNIR.Geometry.Detector import Detector, TubeDetector1D
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt


def test_init():
    GenericDetector = Detector(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    assert(np.all(GenericDetector.position==np.array([0.0,1.0,0.0])))
    assert(np.all(GenericDetector.direction==(1.0,0.0,0.0)))

def test_Generic_plot():
    GenericDetector = Detector(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    plt.ioff()
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    try:
        GenericDetector.plot(ax)
        assert False
    except NotImplementedError:
        assert True



def test_TubeDetector_init():
    TubeDetector = TubeDetector1D(position=(0.0,1.0,0.0),direction=(1.0,0,0),pixels=20,length=0.3,diameter=0.025,split=[0,57,57*2])
    assert(np.all(TubeDetector.position==np.array([0.0,1.0,0.0])))
    assert(np.all(TubeDetector.direction==(1.0,0.0,0.0)))
    assert(TubeDetector.pixels==20)
    assert(TubeDetector.length==0.3)
    assert(TubeDetector.diameter==0.025)
    assert(np.all(TubeDetector.split==np.array([0,57,57*2])))

def test_TubeDetector_pixels():
    TubeDetector = TubeDetector1D(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    try:
        TubeDetector.pixels=0
        assert False
    except AttributeError:
        assert True

def test_TubeDetector_length():
    TubeDetector = TubeDetector1D(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    try:
        TubeDetector.length=-0.1
        assert False
    except AttributeError:
        assert True
    
def test_TubeDetector_diameter():
    TubeDetector = TubeDetector1D(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    try:
        TubeDetector.diameter=-0.1
        assert False
    except AttributeError:
        assert True

def test_TubeDetector_split():
    TubeDetector = TubeDetector1D(position=(0.0,1.0,0.0),direction=(1.0,0,0),pixels=100)
    try:
        TubeDetector.split=-0.1
        assert False
    except AttributeError:
        assert True

    TubeDetector.split=[50,60,100]
    pixelPos = TubeDetector.getPixelPositions()
    assert(len(pixelPos)==2)
    assert(len(pixelPos[0])==10)


def test_TubeDetector1D_plot():
    TubeDetector = TubeDetector1D(position=(0.0,1.0,0.0),direction=(1.0,0,0))
    plt.ioff()
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    TubeDetector.plot(ax)
    

def test_TubeDetector1D_getPixelPositions():
    TubeDetector = TubeDetector1D(position=(1.0,0.0,1.0),direction=(1.0,0,0),length=0.5,pixels=5)
    positions = TubeDetector.getPixelPositions()

    AssumedPositions = np.array([[0.8,0,1],[0.9,0,1],[1.0,0,1],[1.1,0,1],[1.2,0,1]])
    print(positions)
    print(AssumedPositions)
    assert(np.all(AssumedPositions==positions))

    
    


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 28 16:36:22 2023

@author: lass_j
"""


import os,sys,shutil

DIR = r'docs\notebooks'

files = [f.path for f in os.scandir(DIR) if f.is_file() and f.split('.')[-1] == 'ipynb']

for I,f in enumerate(files):
    os.system('jupyter nbconvert --execute --to notebook --inplace '+f)
    print(f'{I} of {len(files)} done')
    

# for folder in ['.cache','.pyteste_cache','__pycache']:
#     shutil.rmtree(os.path.join(DIR,folder))
